package mo.dataprog.utilities;

import java.util.ArrayList;
import java.util.Iterator;

import mo.dataprog.dataStructures.TokenType;
import mo.exceptions.InvalidTokenException;

/**
 *
 *	Class containing methods for various operations
 *	like splitting strings, converting data and other 	
 *  @author Administrator
 *
 */
public final class Utilities
{
	/** Variable used for development purposes - to be removed after finishing of development
	 * 
	 */
	public static final boolean DEBUG = true;
	
	/**
	 * String that contains characters considered as punctuations signs
	 * Note: those signs are not strictly punctuation (like dot, comma, etc.)
	 * 		 This string also contains currency symbols, hyphens etc.
	 */
	public static final String punctuation = ".,;:-_!?'\"/\\&$";

	/**
	 * Method that splits tokens from punctuation signs. Eg: We have a Polish Zip-Code
	 * passed as a string : "03-123", the result of this function is "03 - 123".
	 * After this method, we can easily split string according to "space" characters
	 * @param text
	 * @return
	 */
	public static String SplitTokensFromPunct(String text)
	{
		return SplitTokensFromPunct(new StringBuilder(text)).toString();	
	}
	
	/**
	 * Method that splits tokens from punctuation signs. Eg: We have a Polish Zip-Code
	 * passed as a string : "03-123", the result of this function is "03 - 123".
	 * After this method, we can easily split string according to "space" characters
	 * @param text
	 * @return
	 */
	public static StringBuilder SplitTokensFromPunct(StringBuilder text)
	{
		int punctIndex;
		
		// we iterate through punctuation signs, and check incoming string against it
		for (int i = 0; i < punctuation.length(); i++)
		{
			punctIndex = 0;
			
			// this while loop is for purpose, when a string contains more than one instance
			// of the same punctuation sign
			while (punctIndex >= 0)
			{
			    // since we operate on StringBuffer, we have to operate on strings, not characters
				// so we have to "cut out" single character				
				punctIndex = text.indexOf(punctuation.substring(i,i+1), punctIndex);
				
				// if there is a character
				if (punctIndex >= 0)
				{
					// we have found a punctuation sign in processed string
					
					// if it is not last character in string, and next character is not a whiteSpace,
					// then we put a "space" after the punctuation sign
					if (punctIndex + 1 < text.length() && Character.isWhitespace(text.charAt(punctIndex+1)) == false)
					{	
						text.insert(punctIndex+1," ");					
					}
					
					// if sign befrore token is not a wihtespace, then we insert token
					if (punctIndex - 1 > 0 && Character.isWhitespace(text.charAt(punctIndex-1)) == false)
					{
						// we do not forget about putting the "space" character before the punctuation sign
						text.insert(punctIndex, " ");						
					}
					// we have found punctuation sign at position punctIndex.
					// we have put a white-space at this place
					// so our new punctuation sign has position punctIndex+1
					// since we want to search string from the next position, then
					// we have to add 2 to the index
					punctIndex+=2;
				} // if (punctIndex >=0)
			} // while
			
		}// for
		
		return text;
	}// end method
	
	
	/**
	 * 
	 * @param _text Token to determine its type
	 * @return The Type of token passed as an argument to function
	 * @throws InvalidTokenException 
	 */
	public static  TokenType	determineTokenType(String _text) throws InvalidTokenException
	{
		boolean			isAlphaNum = true;	
		
		// length of the Token - declared as a local variable in order not to call method every
		// time we need to check token's length (and we do it very often)
		int				length =_text.length();
		
		// we do not care about improper tokens
		if (length == 0)
			throw new InvalidTokenException("Token too short!");
	
		// if token is 1-character length, it may be a punctuation sign
		// we check the character against predefined characters recognized as punctuation characters
		if (length == 1 && Utilities.punctuation.indexOf(_text) >= 0)
		{			
			return TokenType.Punct;
		}
		
		// if token starts with "less" sign, then it is HTML tag
		if(_text.startsWith("<"))
		{
			return TokenType.HTML;
		}
		
		// if token is not Punctuation nor HTML tag, then we have to 
		// go deep into hierarchy if tokens and determine to which group
		// does the token belong to
		for (int i = 0; i < length; i++)
		{
			isAlphaNum &= Character.isLetterOrDigit(_text.charAt(i));			
		}
		
		// this means some kind of fatal error - token not assigned to none of three main groups
		// (Punct, HTML, AlphaNum).
		// probably sign from a set that is not applicable for current machine
		if (isAlphaNum == false)
		{
			throw new InvalidTokenException(_text);		
		}
		
		// if we have determined that it really is alpha-numeric char, then we check if 
		// it is possible to tell something more specific about the Token
		if (isAlphaNum == true)
		{
						
			boolean isNumber = true;
			
			// we check if it is a number
			for (int i = 0; i < length; i++)
			{
				isNumber &= Character.isDigit(_text.charAt(i));;				
			}
			if (isNumber == true)
			{
				return TokenType.Number;
			}
			else
			{
				// it is not a number, so it may be alpha token
				boolean isAlpha = true;
				for (int i = 0; i < length; i++)
				{
					isAlpha &= Character.isLetter(_text.charAt(i));;				
				}
				if (isAlpha == false)
				{
					// if it is not a number (determined at previous level)
					// and know we know that it is not Alpha,
					// then it was AlphaNum
					return TokenType.Alphanum;
				}
				else
				{				
					//it is Alpha - we check deeper into hierarchy to check other types					
					boolean isLower = true;
					for (int i = 0; i < length; i++)
					{
						isLower &= Character.isLowerCase(_text.charAt(i));;				
					}
					
					if (isLower == true)
					{
						return TokenType.Lower;
					}
					else
					{
						// if it is not lower, then it may be upper (starting with UpperCase or
						// AllCaps (all characters are uppercase)
						
						// we check if token begins with uppercase  - first letter is upper						
						boolean isUpper = Character.isUpperCase(_text.charAt(0));
						
						
						if (isUpper == false)
						{
							// this is not Upper nor Lower, so it means that upper- and lower- case characters
							// are mixed the token is Alpha - we cannot say anything more specific
							return TokenType.Alpha;
						}
						else
						{
							// it begins with UpperCase
							// if the rest if signs is also uppercase, so it is representatn
							// of AllCaps class
							boolean isAllCaps = true;
							for (int i = 0; i < length; i++)
							{
								isAllCaps &= Character.isUpperCase(_text.charAt(i));;				
							}
							if (isAllCaps == true)
							{
								return TokenType.AllCaps;
							}
							else
							{
								return TokenType.Upper;
							}
						} // else .. isUpper
						
					} // else .. isLower
					
				} // else .. isAlpha
				
			}// else .. isNumber
			
		}// else .. isAlphaNum
		throw new InvalidTokenException("Fatal error during token processing: " + _text);
	}// end method

	
	/**
	 * Method for rounding doubles to specified number of decimial places.
	 * Believe it or not - JAVA does not contain such method in any if its standard libraries
	 * @param toRound
	 * @param numberofDecimalPlaces
	 * @return
	 */
	public static double DEL_Round(double toRound,int numberofDecimalPlaces)
	{
		double factor = Math.pow(10, numberofDecimalPlaces);
		
		return Math.rint(toRound*factor)/factor;
	}

	public static ArrayList<TokenType> GetMoreGeneralTokenTypes(TokenType type)
	{
		ArrayList<TokenType> ret = new ArrayList<TokenType>();
		
		if (type == TokenType.Punct || type == TokenType.Alphanum || type == TokenType.HTML)
			return ret;
		else if (type == TokenType.Alpha || type == TokenType.Number)
		{
			ret.add(TokenType.Alphanum);
		}
		else if (type == TokenType.Upper || type == TokenType.Lower)
		{
			ret.add(TokenType.Alphanum);
			ret.add(TokenType.Alpha);			
		}
		else if (type == TokenType.AllCaps)
		{
			ret.add(TokenType.Alphanum);
			ret.add(TokenType.Alpha);
			ret.add(TokenType.Upper);
		}		
		
		return ret;
		
	}
	
	
	/**
	 * Method for determining size of the longest array of strings
	 * @param examples
	 * @return
	 */
	public static int DetermineLongestElement(ArrayList<ArrayList<String>> examples)
	{
		int longest = -1;
		int temp;
		Iterator<ArrayList<String>> it = examples.iterator();
		ArrayList<String> current;
		
		while (it.hasNext())
		{
			current = it.next();
			temp = current.size();
			
			if (temp > longest)
				longest = temp;			
		}
		
		
		return longest;
	}
	
	public static boolean CheckIfFirstTokenIsMoreGeneral(TokenType first, TokenType second)
	{
		boolean ret;
		
		if (first.equals(second))
			return true;
		
		if (second == TokenType.SpecificType)
			return true;
		
		ArrayList<TokenType> moreGeneralThanSecond = Utilities.GetMoreGeneralTokenTypes(second);
		
		if (moreGeneralThanSecond.contains(first))
			ret = true;
		else
			ret = false;
		
		
		return ret;
		
	}

}
