package mo.dataprog.dataStructures;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Set;

import mo.dataprog.utilities.Utilities;
import mo.exceptions.InvalidTokenException;

/**
 * 
 * @author Administrator
 * Class representing tree of tokens used in DataProg algorithm
 * It contains reference to root node as well as set of methods
 * responsible for basic control of the tree
 */
public class TokenTree
{
	/**
	 * Root of the tree
	 */
	TreeElement root;
	
	
	/**
	 * HashTable for storing token's probabilites - it is based on examples
	 */
	Hashtable<Token, Double> initialProbabilites;
	
	/**
	 * If a line in file containing examples starts with this sign, then we treat it as a comment
	 */
	private final char commentCharacter = '#';
	
	/**
	 * Constructor of the tree - must pass path to file with examples
	 * @param pathToExamples
	 * @throws IOException
	 * @throws InvalidTokenException 
	 */
	public TokenTree(String pathToExamples) throws IOException, InvalidTokenException
	{
		// creation of root
		root = new TreeElement();
		
		// loading of examples
		ArrayList<ArrayList<String>> examples =  this.LoadExamples(pathToExamples);	
		// root of the tree contains all examples
		root.setExamples(examples);
		
		
				
		// TODO: remove after finishing
	/*	if (Utilities.DEBUG)
		{
			Set<Entry<TokenType,Double>> set = initialProbabilites.entrySet();
			
			Iterator <Entry<TokenType,Double>> it =  set.iterator();
			
			while (it.hasNext())
			{
				Entry<TokenType,Double> entry = it.next();
				System.out.println(entry.getKey() + " " + entry.getValue());
			}
		}*/
		
	}

	/**
	 * Method responsible for loading examples into private variable 'examples'
	 * It also fills the HashTable of initial probabilities of tokens
	 * @param pathToExamples
	 * @throws IOException
	 * @throws InvalidTokenException 
	 */
	private ArrayList<ArrayList<String>> LoadExamples(String pathToExamples) throws IOException, InvalidTokenException
	{		
		String currentLine;		
		
		ArrayList<ArrayList<String>> ret = new ArrayList<ArrayList<String>>();
		
		ArrayList<String> curr;
		
		// reader
		BufferedReader bufferedReader = 
			new BufferedReader(
					new FileReader(pathToExamples));
		
		if (bufferedReader != null)
		{			
			// while there is something to read
			while ((currentLine = bufferedReader.readLine()) != null)
			{
				// we create new list
				curr = new ArrayList<String>();
				// we check if the line is marked as a comment
				if (currentLine.charAt(0) == commentCharacter)
				{
					continue;
				}
				else
				{
					// if it is not a comment then we add this as an example (after spliiting punctuatiion Marks from the text)		
					String []tokens = Utilities.SplitTokensFromPunct(currentLine).split(" ");  
					for (String token : tokens)
					{
						curr.add(token);
					}
				}				
				ret.add(curr);
			} // while		
			
			// we do not need stream any more
			bufferedReader.close();			
		} // if				
		
		// we fill probability table
		FillProbabilityHashTable(ret);
		
		return ret;		
	} // end method

	/**
	 * This method is responsible for filling the probability HashTable.
	 * Probabilities are calculated following way - we compute number of occurrences of each
	 * token Type in examples. Then this number is divided by number of all tokens in examples.
	 * @param listWithExamples
	 * @throws InvalidTokenException
	 */
	private void FillProbabilityHashTable(ArrayList<ArrayList<String>> listWithExamples) throws InvalidTokenException
	{
		// we create hash table
		initialProbabilites = new Hashtable<Token, Double>();
		
		InitHashTable();
		
		// local variable for counting how many tokens are in the examples
		double count = 0;
		
		// Iterator for list of examples
		Iterator<ArrayList<String>> it = listWithExamples.iterator();
		
		// temporary variable for current calculations
		ArrayList<String> curr;
		
		Token currToken;
		
		
		while (it.hasNext())
		{
			curr = it.next();
			
			// iterator for current variable
			Iterator<String> itCurr = curr.iterator();
			
			while (itCurr.hasNext())
			{
				count++;
				String actStr = new String(itCurr.next()); 
				
				TokenType type = Utilities.determineTokenType(actStr);
				Double currValue;
				// at first we count how many elements of which type are in the examples
				// calculation of probability will be done later
				
				// add more general types
				ArrayList<TokenType> moreGeneralTokenTypes = Utilities.GetMoreGeneralTokenTypes(type);
				for (TokenType t : moreGeneralTokenTypes)
				{
					currToken = new Token(t);
					currValue =  initialProbabilites.get(currToken);				
					currValue++;
					initialProbabilites.put(currToken, currValue);	
				}
				
				// add the type itself
				currToken = new Token(type);
				currValue =  initialProbabilites.get(currToken);				
				currValue++;
				initialProbabilites.put(currToken, currValue);
				
				// add specific type
				currToken = new Token(actStr);
				currValue = initialProbabilites.get(currToken);
				if (currValue == null)
				{
					currValue = 0d;
				}
				currValue++;		
				initialProbabilites.put(currToken, currValue);
			}
		} // while (it.hasNext())		
		
		
		
		// at this moment we have probability HashTable filled with tokenTypes and corresponding number of occurencies
		// we also have total number of tokens. So for each element we can compute the probability of occurence of tokens
		
		Set<Entry<Token,Double>> entrySet = initialProbabilites.entrySet();
		
		Entry<Token,Double> currentEntry;
		Iterator<Entry<Token,Double>> itEntry = entrySet.iterator();
		while (itEntry.hasNext())
		{
			
			currentEntry = itEntry.next();
		 
					
			currentEntry.setValue(currentEntry.getValue()/count);//Utilities.Round(currentEntry.getValue()/count,2));
		//	System.out.println(currentEntry.getKey().toString() + " " + currentEntry.getValue());
		}
	}

	
	private void InitHashTable()
	{
		initialProbabilites.put(new Token(TokenType.AllCaps), 0d);
		initialProbabilites.put(new Token(TokenType.Alpha), 0d);
		initialProbabilites.put(new Token(TokenType.Alphanum), 0d);
		initialProbabilites.put(new Token(TokenType.HTML), 0d);
		initialProbabilites.put(new Token(TokenType.Lower), 0d);
		initialProbabilites.put(new Token(TokenType.Number), 0d);
		initialProbabilites.put(new Token(TokenType.Punct), 0d);		
		initialProbabilites.put(new Token(TokenType.Upper), 0d);		
	}

	/**
	 * Return the root of the tree
	 * @return
	 */
	public TreeElement getRoot()
	{
		return root;
	}

	public Hashtable<Token, Double> getInitialProbabilites()
	{
		return initialProbabilites;
	}
	
	
	
}
