package semfly.agents;

import java.util.PriorityQueue;

import semfly.agents.behaviours.CoordinatorGUIReceiver;
import semfly.agents.behaviours.CoordinatorReceiver;
import semfly.agents.behaviours.CoordinatorScheduler;
import semfly.agents.behaviours.DeleteMyself;
import semfly.exceptions.ExceptionManager;
import semfly.globals.DFAdapter;
import semfly.onto.jade.IndexToken;
import semfly.onto.jade.TokensOntology;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.core.AID;
import jade.core.Agent;
import jade.domain.DFService;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.core.behaviours.ThreadedBehaviourFactory;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 * 
 * This class represents Coordinator Agent who plays vital role in 
 * Content Collection Subsystem. He is responsible for:
 * 1. starting Indexing Agents at CCS start-up
 * 2. starting Wrapper Agents at CCS start-up and starting them in appropriate
 * 		time (in appropriate intervales)
 * 
 * This Agent also acts as a communication channel between Wrapper Agents and
 * Indexing Agents. 
 * Wrapper Agents send token (IndexToken) with metadata about 
 * some resource. Coordinator saves token in its TPQ (Tokens Priority Queue, each
 * token has its priority).
 * Indexing Agent sends a request for any token (IndexToken) to save 
 * in a database. Coordinator gets a token with the highest priority from TPQ 
 * and sends it to the Indexing Agent, whose the next task is to save the token 
 * in DB.
 */
public class CoordinatorAgent extends Agent implements ITPQOwner
{ 
    /**
	 * Describes type of the service an agent does.
	 */
	public final static String ServiceType = "CoordinatingService";	
	
	/**
	 * Timeout for receiving messages in Contract-Net-Protocol.
	 */
	private final int ContractTimeout = 10000;
	
	/**
	 * Agent's codex.
	 */
	private Codec codec = new SLCodec();
	
	/**
	 * Agent's ontology.
	 */
	private Ontology ontology = TokensOntology.getInstance();
	
	/**
	 * Tokens Priority Queue (TPQ). Collection of tokens received from
	 * Wrapper Agents.
	 */
	private PriorityQueue tokensPQ;
	
	/**
	 * Behaviour responsible for receiving ACL messages from Wrapper Agents
	 * and Indexing Agents.
	 */
	private CoordinatorReceiver receiverBehaviour;
	
	/**
	 * Behaviour responsible for starting:
	 * 1. Indexing Agents - at system start-up
	 * 2. Wrapper Agents - at system start-up and after appropriate intervals
	 */
	private CoordinatorScheduler schedulerBehaviour;
	
	/**
	 * Constructor.
	 */
	public CoordinatorAgent() 
	{
		super();
	}
	
	/**
	 * Agent initialization.
	 * Creates Tokens Priority Queue.
	 * Registers Agent's language and ontology.
	 * Registers Agent in DF service.
	 * Adds Coordinator behaviours: CoordinatorReceiver and CoordinatorScheduler.
	 * 
	 * @see jade.core.Agent#setup()
	 */
	protected void setup() 
	{
	    tokensPQ = new PriorityQueue();
	    
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(ontology);
		
		ServiceDescription sd  = new ServiceDescription();
        sd.setType(ServiceType);
        sd.setName(getLocalName());
        register(sd);
		
        addReceiverBehaviour();
        addSchedulerBehaviour();
        addGUIReceiverBehaviour();
	}
	
	/**
	 * Adds CoordinatorReceiver behaviour to Agent's behaviours.
	 * This behaviour is to be executed in dedicated thread.
	 */
	protected void addReceiverBehaviour()
	{
	    MessageTemplate template = prepareIncommingMsgTemplate();
		
	    ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
	    receiverBehaviour = new CoordinatorReceiver(this, template);
		addBehaviour(tbf.wrap(receiverBehaviour));
	}
	
	/**
	 * Prepares template for incomming messages from Wrapper Agents 
	 * and Indexing Agents.
	 * 
	 * @return Message template.
	 */
	protected MessageTemplate prepareIncommingMsgTemplate()
	{
		MessageTemplate template = MessageTemplate.and(
	  		MessageTemplate.MatchProtocol(
	  			FIPANames.InteractionProtocol.FIPA_CONTRACT_NET),
					MessageTemplate.MatchPerformative(ACLMessage.CFP));
		
		template = MessageTemplate.and(template, 
						MessageTemplate.and(MessageTemplate.
							MatchLanguage(codec.getName()),
							MessageTemplate.MatchOntology(ontology.getName())));
		
		return template;
	}
	
	/**
	 * Adds CoordinatorScheduler behaviour to Agent's behaviours.
	 * This behaviour is to be executed in dedicated thread.
	 */
	protected void addSchedulerBehaviour()
	{
	    ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
	    
	    schedulerBehaviour = new CoordinatorScheduler(this);
		addBehaviour(tbf.wrap(schedulerBehaviour));
	}
	
	/**
	 * Adds CoordinatorGUIReceiver behaviour to Agent's behaviours.
	 * This behaviour is to be executed in a dedicated thread.
	 */
	protected void addGUIReceiverBehaviour()
	{
	    MessageTemplate template = prepareGUIMsgTemplate();
		
	    ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
		addBehaviour(tbf.wrap(new CoordinatorGUIReceiver(this, template)));
	}
	
	/**
	 * Prepares template for incomming messages from GUI Agent.
	 * 
	 * @return Message template.
	 */
	protected MessageTemplate prepareGUIMsgTemplate()
	{
	    MessageTemplate template = MessageTemplate.and(
		  		MessageTemplate.MatchProtocol(
		  			FIPANames.InteractionProtocol.FIPA_REQUEST),
						MessageTemplate.MatchPerformative(ACLMessage.REQUEST));
		
		/*
		template = MessageTemplate.and(template, 
						MessageTemplate.and(MessageTemplate.
							MatchLanguage(codec.getName()),
							MessageTemplate.MatchOntology(ontology.getName())));
		*/
		return template;
	}
	
	/**
	 * Agent's clean-up. Deregisters Agent from DF service.
	 * 
	 * @see jade.core.Agent#takeDown()
	 */
	protected void takeDown() 
	{
	    try 
		{
	        DFService.deregister(this); 
		}
        catch (Exception ex) 
		{
            ExceptionManager.publish(ex);
        }
	}
	
	/**
	 * Commits suicide.
	 */
	public void commitSuicide()
	{
	    ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
	    
	    Thread thread = tbf.getThread(receiverBehaviour);
	    if (thread != null) thread.interrupt();
	    
	    thread = tbf.getThread(schedulerBehaviour);
	    if (thread != null) thread.interrupt();
	    
	    addBehaviour(new DeleteMyself());
	}
	
	/**
	 * Registeres an agent in DF service by means of DFAdapter class.
	 * @param sd Description of a service Agent wants to register
	 */
	protected void register(ServiceDescription sd)
	{
	    DFAdapter adapter = DFAdapter.getInstance();
		adapter.register(this, sd);
	}
	
	/**
	 * Returns Agent's codec (language).
	 * @return Agent's codec.
	 */
    public Codec getCodec()
    {
        return codec;
    }
    
    /**
     * Returns Agent's ontology.
     * @return Agent's ontology.
     */
    public Ontology getOntology()
    {
        return ontology;
    }
	
    /**
     * Retrieves and removes a token with the highest priority from TPQ.
     * @return Token with the highest priority.
     */
    public IndexToken poolIndexToken()
    {
        synchronized (tokensPQ)
        {
            return (IndexToken)tokensPQ.poll();
        }
    }
    
    /**
     * Returns number of tokens (IndexToken) in TPQ.
     * @return Number of tokens.
     */
    public int getNumberOfTokens()
    {
        synchronized (tokensPQ)
        {
            return tokensPQ.size();
        }
    }
    
    /**
     * Checks if there are any tokens in TPQ.
     * @return True - there is at least one token in TPQ, otherwise False.
     */
    public boolean hasIndexTokens()
    {
        synchronized (tokensPQ)
        {
            return tokensPQ.size() > 0;
        }
    }
    
    /**
     * Adds the token given to TPQ.
     * @param token Token to be added.
     */
	public void addIndexToken2TPQ(IndexToken token)
	{
	    assert token != null;
	    
	    synchronized (tokensPQ)
        {
	        tokensPQ.add(token);
        }
	}
	
	/**
	 * Searches all agents based on the service type given by means of DFAdapter
	 * class.
	 * @param service Service type.
	 * @return AIDs of found agents of the service given, null if no agent 
	 * has been found or an exception has occured.
	 */
	protected AID[] searchDF(String service)
	{
	    DFAdapter searcher = DFAdapter.getInstance();
		
		return searcher.searchDF(this, service);
	}
}
