package semfly.agents;

import java.util.Date;
import semfly.agents.behaviours.BusinessBehaviour;
import semfly.agents.behaviours.DeleteMyself;
import semfly.agents.behaviours.IndexingAgentRequestSender;
import semfly.agents.behaviours.OrdersReceiver;
import semfly.exceptions.ExceptionManager;
import semfly.globals.DFAdapter;
import jade.core.AID;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.domain.FIPANames;
import jade.domain.DFService;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import semfly.onto.jade.RequestToken;
import semfly.onto.jade.TokensOntology;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.ThreadedBehaviourFactory;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 * 
 * This class represents an Agent who is responsible for saving metadata 
 * descriptions about some resources in a RDF database.
 */
public class IndexingAgent extends Agent
{
	/**
	 * Describes type of the service an agent does.
	 */
	public final static String ServiceType = "IndexingService";	
	
	/**
	 * Timeout for receiving messages in Contract-Net-Protocol.
	 */
	private final int ContractTimeout = 1000;
	
	/**
	 * Agent sleeps for this time before sending next token request.
	 */
	private final int BreakTime = 10000;
	
	/**
	 * Agent's codec (language).
	 */
	private Codec codec = new SLCodec();
	
	/**
	 * Agent's ontology.
	 */
	private Ontology ontology = TokensOntology.getInstance();
	
	/**
	 * Agent's main bahaviour. This is a composite behaviour.
	 */
	private BusinessBehaviour busBehaviour;
	
	private OrdersReceiver ordersBehaviour;
	
	private boolean commitSuicide = false;
	
	/**
	 * Constructor.
	 */
	public IndexingAgent()
	{
		super();
	}
	
	/**
	 * Agent initialization.
	 * Registers Agent's language and ontology.
	 * Registers Agent in DF service.
	 * Creates Agent's main behaviour (busBehaviour) and 
	 * adds to it IndexingAgentRequestSender behaviour as a subbahaviour.
	 */
	protected void setup()
	{
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(ontology);
		
		ServiceDescription sd  = new ServiceDescription();
        sd.setType(ServiceType);
        sd.setName(getLocalName());
        register(sd);
        
        ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
        ordersBehaviour = new OrdersReceiver(this);
		addBehaviour(tbf.wrap(ordersBehaviour));
        
	  	busBehaviour = new BusinessBehaviour(this);
	  	ACLMessage tokenRequest = prepareACLMessage();
	  	if (tokenRequest != null)
	  	{
		  	busBehaviour.addSubBehaviour(
		  	        new IndexingAgentRequestSender(this, tokenRequest));
	  	}
	  	else addRequestSenderBehaviour();
	  	addBehaviour(busBehaviour);
	}
	
	/**
	 * Does cleaning before agent death.
	 * Deregisters Agent from DF service.
	 */
	protected void takeDown()
	{
		 try 
		 { 
		 	DFService.deregister(this); 
		 }
         catch (Exception ex) 
		 {
         	ExceptionManager.publish(ex);
         }
	}
	
	/**
	 * Registeres an agent in DF service by means of DFAdapter class.
	 * @param sd Description of a service Agent wants to register
	 */
	private void register(ServiceDescription sd)
	{
		DFAdapter adapter = DFAdapter.getInstance();
		adapter.register(this, sd);
	}
	
	/**
	 * Prepares an ACL message that will be sent to Coordinator Agent. 
	 * This message includes request for a token (IndexToken).
	 * @return Created ACL message.
	 */
	protected ACLMessage prepareACLMessage()
	{
		ACLMessage msg = new ACLMessage(ACLMessage.CFP);
  		msg.setLanguage(codec.getName());
  		msg.setOntology(ontology.getName());
  		
  		AID[] coordinators = searchDF(CoordinatorAgent.ServiceType);
  		
  		if (coordinators != null)
  		{
  		  for (int i = 0; i < coordinators.length; i++)
	  			msg.addReceiver(coordinators[i]);
	  		
	  		msg.setProtocol(FIPANames.InteractionProtocol.FIPA_CONTRACT_NET);
	  		msg.setReplyByDate(
	  				new Date(System.currentTimeMillis() + ContractTimeout));
	  		
	  		Action contentAction = createAgentAction();
	  		
	  		try 
			{
	  			getContentManager().fillContent(msg, contentAction);
	  		}
	  		catch (Codec.CodecException ex) 
			{
	  		    ExceptionManager.publish(ex);
	  		}
	  		catch (OntologyException ex) 
			{
	  		    ExceptionManager.publish(ex);
	  		}
	  	
	  		return msg;
  		}
  		else
  		{
  			return null;
  		}
	}
	
	/**
	 * Creates RequestToken action.
	 * Coordinator will be asked to do this action.
	 * @return Created action.
	 */
	private Action createAgentAction()
	{
		Action contentAction = new Action();
  		contentAction.setAction(new RequestToken());
  		contentAction.setActor(getAID());
  		
		return contentAction;
	}
	
	/**
	 * Commits suicide by removing all behaviours and calling doDelete().
	 */
	public void commitSuicide()
	{
	    commitSuicide = true;
	    
	    addBehaviour(new DeleteMyself());
	}
	
	/**
	 * Searches all agents based on the service type given by means of DFAdapter
	 * class.
	 * @param service Service type.
	 * @return AIDs of found agents of the service given, null if no agent 
	 * has been found or an exception has occured.
	 */
	protected AID[] searchDF(String service)
	{
	    DFAdapter searcher = DFAdapter.getInstance();
		
		return searcher.searchDF(this, service);
	}
	
	/**
	 * Adds subbehaviour to Agent's main behaviour.
	 * @param behaviour Behaviour to be added.
	 */
	public void addBusinessSubBehaviour(Behaviour behaviour)
	{
	    assert behaviour != null;
	    
	    if (commitSuicide) return;
	    busBehaviour.addSubBehaviour(behaviour);
	}
	
	/**
	 * Adds subbehaviour to Agent's main behaviour that is
	 * responsible for sending token request to Coordinator Agent.
	 *
	 */
	public void addRequestSenderBehaviour()
	{
	    if (commitSuicide) return;
	    
        try { Thread.sleep(BreakTime); } catch (Exception ex){}
        
	    ACLMessage tokenRequest = prepareACLMessage();
	    
	    if (tokenRequest != null)
	    {
		    busBehaviour.addSubBehaviour(
		            new IndexingAgentRequestSender(this, tokenRequest));
	    }
	    else addRequestSenderBehaviour();
	}
	
	/**
	 * Returns Agent's language.
	 * @return Agent's language.
	 */
	public String getLanguageName()
	{
		assert codec != null;
		
		return codec.getName();
	}
	
	/**
	 * Returns Agent's ontology.
	 * @return Agent's ontology.
	 */
	public String getOntologyName()
	{
		assert ontology != null;
		
		return ontology.getName();
	}
}
