package semfly.agents;

import jade.core.AID;
import jade.gui.GuiAgent;
import jade.gui.GuiEvent;
import jade.wrapper.AgentController;
import jade.wrapper.ControllerException;
import jade.wrapper.PlatformController;
import semfly.globals.CCSVocabulary;
import semfly.globals.DFAdapter;
import semfly.gui.GUIVocabulary;
import semfly.gui.SystemGUI;
import semfly.agents.behaviours.GUIAgentSender;
import semfly.exceptions.ExceptionManager;
import jade.lang.acl.ACLMessage;
import jade.domain.FIPANames.InteractionProtocol;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 * 
 * This class represents a GUI Agent.
 */
public class SystemGUIAgent extends GuiAgent implements GUIVocabulary, 
	CCSVocabulary
{
    /**
     * Name of the Coordinator class (is used while creating Coordinator Agent).
     */
    private final static String CoordinatorClass = 
        "semfly.agents.CoordinatorAgent";
    
    /**
     * Name of the Coordinator Agent.
     */
    private final static String CoordinatorName = "CoordinatorAgent";
    
    private SystemGUI systemGUI;
    
    private boolean systemStarted = false;

    /**
     * Tells if Coordinator and all Indexing Agents and Wrapper Agents died.
     */
    private boolean canDie = false;

	protected void setup() 
	{
	    systemGUI = new SystemGUI(this);
	    systemGUI.showGUI();
	}

	protected void takeDown() 
	{
	    
        if (systemGUI!=null) 
        {
            systemGUI.setVisible(false);
            systemGUI.dispose();
        }
        System.out.println("Content Collection Subsystem is shutted down.");
    }
	
	/**
	 * Processes events from GUI (SystemGUI class).
	 * 
	 * @see jade.gui.GuiAgent#onGuiEvent(jade.gui.GuiEvent)
	 */
	protected void onGuiEvent(GuiEvent event)
	{
	    int command = event.getType();
	    switch (command)
	    {
	        case START_CCS:
	            startCCS();
	            break;
	        case SHUTDOWN_CCS:
	            shutDownCCS();
	            break;
	    }
	}
	
	private void startCCS()
	{
	    if (systemStarted) return;
	    
	    PlatformController platform = this.getContainerController();
	    try
		{
			AgentController agent;
			agent = platform.createNewAgent(CoordinatorName, 
			        CoordinatorClass, null);
			agent.start();
		}
		catch (ControllerException ex)
		{
			ExceptionManager.publish(ex);
		}
		
		systemStarted = true;
		System.out.println("Content Collection Subsystem has just been started.");
	    
	}
	
	/**
	 * Orders Coordinator to die and in consequence Coordinator is to
	 * kill all working Indexing Agents and Wrapper Agents.
	 */
	private void shutDownCCS()
	{
	    System.out.println("Content Collection Subsystem started shutting down"
	            + " process.");
	    sendOrder2Die();
	}
	
	/**
	 * Makes this agent dead and exits the CCS.
	 */
	public void commitSuicide()
	{
	    doDelete();
	}
	
	/**
	 * Sends "order to die" to one of agents of CoordinatorAgent type.
	 * If any exception is thrown or there is no receivers (Coordinators), 
	 * this agent dies and shuts down the CCS.
	 */
	private void sendOrder2Die()
	{
	    ACLMessage msg = null;
		try
		{
		     msg = prepareOrder2DieMessage();
		     addSenderBehaviour(msg);
		}
		catch (Exception ex)
		{
		    ExceptionManager.publish(getName(), ex, 
		            "ACL message preparation failed");
		}
		
		if (msg == null) commitSuicide();
	}
	
	/**
	 * Prepares message being an "order to die" for one of agents of
	 * type CoordinatorAgent. Searches for Coordinators and calls
	 * <code>fillDieMsgContent()</code> in order to fill the content of the
	 * message.
	 * 
	 * @return Prepared message.
	 */
	private ACLMessage prepareOrder2DieMessage()
	{
	    DFAdapter searcher = DFAdapter.getInstance();
	    AID[] coordinators = 
	        searcher.searchDF(this, CoordinatorAgent.ServiceType);
  		
  		if (coordinators != null && coordinators.length > 0)
  		{
  		    return fillDieMsgContent(coordinators);
	  	}
  		
  		return null;
	}
	
	/**
	 * Fills content of "order to die" message. 
	 * @param receivers List of receivers.
	 * @return Filled ACL message.
	 */
	private ACLMessage fillDieMsgContent(AID[] receivers)
	{
	    assert receivers != null;
	    
	    ACLMessage request = new ACLMessage(ACLMessage.REQUEST);
	    request.setProtocol(InteractionProtocol.FIPA_REQUEST);
	    for (int i = 0; i < receivers.length; i++)
	        request.addReceiver(receivers[i]);
	    
	    request.setContent(ORDER_2_DIE);
	    
	    return request;
	}
	
	/**
	 * Adds a behaviour responsible for sending the message given.
	 * @param msg Message to be sent.
	 */
	public void addSenderBehaviour(ACLMessage msg)
	{
	    addBehaviour(new GUIAgentSender(this, msg));
	}
}
