package semfly.agents;

import semfly.agents.behaviours.BusinessBehaviour;
import semfly.agents.behaviours.DeleteMyself;
import semfly.agents.behaviours.OrdersReceiver;
import semfly.exceptions.ExceptionManager;
import semfly.globals.DFAdapter;
import jade.core.Agent;
import jade.core.behaviours.SequentialBehaviour;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.ThreadedBehaviourFactory;
import jade.content.lang.sl.SLCodec;
import jade.content.lang.Codec;
import jade.content.onto.Ontology;
import jade.domain.DFService;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import semfly.onto.jade.TokensOntology;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class represents basic class for all Wrapper Agents.
 * 
 * This Agent creates token (called IndexToken since now) based upon gathered 
 * data, a token includes some administrative info too 
 * (created/modified/modification-try date, created/modified/modification-try by 
 * - agent name, token priority). 
 * Then it sends request for taking care of the IndexToken to the 
 * Coordinator (Contract Net Protocol). After successfully sending a token,
 * Agent dies.
 */
public abstract class WrapperAgent extends Agent 
{
    /**
     * Type of the service an Agent does.
     */
    public static final String ServiceType = "WrapperService";
    
    /**
     * Agent's main behaviour.
     */
    private SequentialBehaviour baseBehaviour;
    
    /**
     * Agent's main business behaviour. This is a composite behaviour.
     */
	private BusinessBehaviour businessBehaviour;
	
	/**
	 * Agent's codec (language).
	 */
	private Codec codec = new SLCodec();
	
	/**
	 * Agent's ontology.
	 */
	private Ontology ontology = TokensOntology.getInstance();
	
	private boolean commitSuicide = false;
	
	private int numberOfVisitedPages = 0;
	
	/**
	 * Constructor.
	 */
	public WrapperAgent() 
	{
	    super();
	}
	
	/**
	 * Agent initialization.
	 * Registers agent language and ontology.
	 * Registers Agent in DF service.
	 * Adds suquential behaviour that consist of the following behaviours:
	 * BusinessBehaviour (SequentialBehaviour) - responsible for "business work"
	 * and
	 * DeleteMyself - responsible for deleting Agent.
	 * 
	 * @see jade.core.Agent#setup()
	 */
	protected void setup() 
	{
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(ontology);
		
		ServiceDescription sd  = new ServiceDescription();
        sd.setType(ServiceType);
        sd.setName(getLocalName());
        register(sd);
		
		baseBehaviour = new SequentialBehaviour(this);
		businessBehaviour = new BusinessBehaviour(this);
		
		ThreadedBehaviourFactory tbf = new ThreadedBehaviourFactory();
		addBehaviour(tbf.wrap(new OrdersReceiver(this)));
		
		baseBehaviour.addSubBehaviour(businessBehaviour);
		baseBehaviour.addSubBehaviour(new DeleteMyself());
		
		addBehaviour(baseBehaviour);
	}
	
	/**
	 * Registeres an agent in DF service by means of DFAdapter class.
	 * @param sd Description of a service Agent wants to register
	 */
	private void register(ServiceDescription sd)
	{
		DFAdapter adapter = DFAdapter.getInstance();
		adapter.register(this, sd);
	}
	
	/**
	 * Does cleaning before agent death.
	 * Deregisters Agent from DF service.
	 */
	protected void takeDown()
	{
		 try 
		 { 
		 	DFService.deregister(this);
		 }
         catch (Exception ex) 
		 {
         	ExceptionManager.publish(ex);
         }
	}
	
	/**
	 * Adds bahaviour to the agent business behaviours.
	 * @param behaviour The behaviour to be added.
	 */
	public void addBusinessSubBehaviour(Behaviour behaviour)
	{
		assert businessBehaviour != null;
		assert behaviour != null;
		
		if (commitSuicide) return;
		businessBehaviour.addSubBehaviour(behaviour);
	}
	
	/**
	 * Commits suicide by removing all behaviours and calling doDelete().
	 */
	public void commitSuicide()
	{
	    commitSuicide = true;
	    
	    addBehaviour(new DeleteMyself());
	}
	
	/**
	 * Returns Agent's language.
	 * @return Agent's language.
	 */
	public String getLanguageName()
	{
		assert codec != null;
		
		return codec.getName();
	}
	
	/**
	 * Returns Agent's ontology.
	 * @return Agent ontology.
	 */
	public String getOntologyName()
	{
		assert ontology != null;
		
		return ontology.getName();
	}
	
	public int getNumberOfVisitedPages()
	{
	    return numberOfVisitedPages;
	}
	
	/**
	 * Increments number of visited pages.
	 */
	public void incNumberOfVisitedPages()
	{
	    synchronized(this)
	    {
	        numberOfVisitedPages++;
	    }
	}
}
