package semfly.agents.behaviours;

import jade.core.Agent;
import semfly.agents.WrapperAgent;
import semfly.dto.ResourceDocument;
import semfly.dto.WebPageDocument;
import semfly.exceptions.ExceptionManager;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This behaviour loads one page of 8 from the "code site" 
 * (www.world-airport-codes.com, I use "Airport codes and details ordered 
 * by country code" section). 
 * After loading, codes are extracted from the page and given as an input 
 * parameter to AirportInfoLoader behaviour. If not all pages (8) are loaded 
 * a new AirportsCodesLoader behaviour is added and the next page is going to be 
 * taken.
 */
public final class AirportsCodesLoader extends ResourceLoader
{
    /**
     * This class is responsible for "business logic" within this behaviour.
     */
    private AirportsCodesLoaderBL airLoader;

	/**
	 * Constructor. Sets current codes' page index on zero.
	 */
	private AirportsCodesLoader()
	{
		super();
		airLoader = new AirportsCodesLoaderBL(myAgent);
	}

	/**
	 * Constructor. Sets current codes' page index on zero.
	 * @param a Agent - owner of this behaviour.
	 */
	public AirportsCodesLoader(Agent a)
	{
		super(a);
		airLoader = new AirportsCodesLoaderBL(myAgent);
	}

	/**
	 * Main method of the Behaviour class. Starts loading airports' codes.
	 */
	public void action() 
	{
	    try
		{
			loadResource();
		}
		catch (Exception ex)
		{
			ExceptionManager.publish(ex);
		}
	}
	
	/**
	 * Overriden method. 
	 * If not all pages (8) are loaded a new AirportsCodesLoader behaviour 
	 * is added and the next page is going to be taken.
	 */
	public int onEnd()
	{
	    super.onEnd();
		if (airLoader.morePages2Load())
		{
			airLoader.incCurrentPageIdx();
			((WrapperAgent)myAgent).addBusinessSubBehaviour(this);
		}
		
		return 0;
	}
	
	/** 
	 * Loads page with airports' codes.
	 * @return Loaded page.
	 * @throws WebSourceLoadException Problems with loading a web page.
	 * 
	 * @see semfly.agents.behaviours.ResourceLoader#loadResource()
	 */
	protected ResourceDocument loadResource() throws WebSourceLoadException
	{
		
		WebPageDocument codes = airLoader.loadAirportCodes();
		
		try
		{
			((WrapperAgent)myAgent).addBusinessSubBehaviour(
					new AirportsInfoLoader(myAgent, 
					        airLoader.extractAirportCodes(
					                codes.getWebDocument())));
		}
		catch (WebSourceParseException ex)
		{
		    ExceptionManager.publish(ex);
		}
		
		return codes;
	}
}
