package semfly.agents.behaviours;

import jade.core.Agent;

import java.util.List;

import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import semfly.dto.WebPageDocument;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for AirportsCodesLoader behaviour.
 * 
 * This behaviour loads one page of 8 from the "code site" 
 * (www.world-airport-codes.com, I use "Airport codes and details ordered 
 * by country code" section). 
 * After loading, codes are extracted from the page and given as an input 
 * parameter to AirportInfoLoader behaviour. If not all pages (8) are loaded 
 * a new AirportsCodesLoader behaviour is added and the next page is going to be 
 * taken.
 */
public class AirportsCodesLoaderBL
{
    /**
	 * Base URI address of the web site with airports codes.
	 */
	private String codesBaseURI;
	
	/**
	 * Array of endings for the base URI address 
	 * of the web site with airports codes.
	 */
	private String[] codesEndURI;
	
	/**
	 * Number of pages with airports' codes.
	 */
	private int NoCodePages;
	
	/**
	 * Current index of the "end URI" of a page with airport codes 
	 * in codesEndURI array.
	 */
	private int curCodesPageIdx;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	/**
	 * Constructor. Sets current codes' page index on zero.
	 */
	public AirportsCodesLoaderBL(Agent agent)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    
	    init();
		myAgent = agent;
	}
	
	private void init()
	{
	    curCodesPageIdx = 0;
	    
	    codesBaseURI = "http://www.world-airport-codes.com/alphabetical/country-code/";
		codesEndURI = new String[]{"0-49.html", "50-99.html", "100-149.html",
				"150-199.html", "200-249.html", "250-299.html", "300-349.html"};
		
		NoCodePages = codesEndURI.length;
	}
	
	/**
	 * Loads info about aiport codes from the page of:
	 * "codesBaseURI + codesEndURI[curCodesPageIdx]" URL address. 
	 * @return Loaded web page with info about airport codes.
	 * @throws WebSourceLoadException
	 */
	public WebPageDocument loadAirportCodes() throws WebSourceLoadException
	{
	    return new WebPageDocument(loadAirportCodes(codesBaseURI + 
									codesEndURI[curCodesPageIdx]));
	}
	
	/**
	 * Loads info about airport codes.
	 * @param pageURI URI address of a page to load.
	 * @return Loaded web page with info about airport codes.
	 */
	private Document loadAirportCodes(String pageURI) 
		throws WebSourceLoadException
	{	
		WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		GetMethod httpMethod = new GetMethod(pageURI);
		try
		{
			doc = webLoader.loadWebSiteHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadAirportCodes() failed", ex, pageURI);
		}
		
		return doc;
	}

	/**
	 * Extracts airport codes for the page with codes given.
	 * @param codesDoc Page with codes.
	 * @return Array of airports codes.
	 * @throws WebSourceParseException Parsing exception.
	 */
	public String[] extractAirportCodes(Document codesDoc) 
		throws WebSourceParseException
	{
	    if (codesDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
		try
		{
			DOMXPath xpath = new DOMXPath("/HTML/BODY/TABLE[1]/TR[1]/TD[2]/" +
					"TABLE[1]/TR[2]/TD[1]/TABLE[1]/TR[2]/TD[1]/TABLE[2]" +
					"/TR[@class='one' | @class='two']/TD[6]/text()");
			
			List nodes = xpath.selectNodes(codesDoc);
			if (nodes.size() <= 0)
			{
			    throw new WebSourceParseException(myAgent.getAID(), 
				        "extractAirportCodes() failed",  
				        codesBaseURI + codesEndURI[curCodesPageIdx]);
			}
			String[] codes = new String[nodes.size()];
			for (int i = 0; i < codes.length; i++)
				codes[i] = ((Node)nodes.get(i)).getNodeValue();
			
			return codes;
		}
		catch (JaxenException ex)
		{
			throw new WebSourceParseException(myAgent.getAID(), 
			        "extractAirportCodes() failed", ex, 
			        codesBaseURI + codesEndURI[curCodesPageIdx]);
		}
	}
	
	/**
	 * Checks whether there are more pages with airports codes to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curCodesPageIdx + 1 < codesEndURI.length;
	}
	
	/**
	 * Increments "curCodesPageIdx" variable.
	 */
	public void incCurrentPageIdx()
	{
	    curCodesPageIdx += 1;
	}
	
	/**
	 * Sets base URI for pages with airports' codes. This method is used only
	 * for testing purposes.
	 * 
	 * @param baseURI URI to set.
	 */
	public void setCodeBaseURI(String baseURI)
	{
	    codesBaseURI = baseURI;
	}
	
	/**
	 * Sets an array of endings for the base URI address 
	 * of the web site with airports codes. It also sets "NoCodePages" variable.
	 * This method is used only for testing purposes.
	 * 
	 * @param endURIs
	 */
	public void setCodesEndURI(String[] endURIs)
	{
	    codesEndURI = endURIs;
		NoCodePages = codesEndURI.length;
	}
}
