package semfly.agents.behaviours;

import jade.core.Agent;
import org.w3c.dom.Document;
import semfly.agents.WrapperAgent;
import semfly.exceptions.ExceptionManager;
import semfly.exceptions.WebSourceLoadException;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This behaviour loads necessary details for some of the given codes. 
 * The important thing is that this behaviour does not load details for all 
 * the codes given at once! 
 * After loading some info from "details site" 
 * (http://www.airrouting.com/content/airportloc.html) for e.g. 5 codes 
 * it gives loaded resources (as DOM objects) to the AirportsInfoProcessor. 
 * And if not all details are loaded (for all the given codes) 
 * a new AirportsInfoLoader bahaviour is added and the next details are going to 
 * be loaded.
 */
public final class AirportsInfoLoader extends ResourceLoader
{
    /**
     * This class is responsible for "business logic" within this behaviour.
     */
    private AirportsInfoLoaderBL airLoader = null;

	/**
	 * Constructor.
	 * @param owner Agent - owner of this behaviour.
	 */
	private AirportsInfoLoader(Agent owner)
	{
		super(owner);
	}
	
	/**
	 * Constructor. Sets airports codes for which all necessary details should 
	 * be loaded (airCodes param) and currend code index on zero.
	 * @param owner Agent - owner of this behaviour.
	 * @param airCodes Airports codes for which all necessary details should 
	 * be loaded.
	 */
	public AirportsInfoLoader(Agent owner, String[] airCodes)
	{
	    super(owner);
	    
	    airLoader = new AirportsInfoLoaderBL(owner, airCodes);
	}
	
	/** 
	 * Main mehtod of any Behaviour class.
	 * Starts loading info about airports and adds beahviour 
	 * (AirportsInfoProcessor) responsible for processing loaded data.
	 * 
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action() 
	{
		try
		{
			Document[] airportsDetails = airLoader.loadResources();
			((WrapperAgent)myAgent).addBusinessSubBehaviour(
					new AirportsInfoProcessor(myAgent, airportsDetails));
		}
		catch (WebSourceLoadException ex)
		{
			ExceptionManager.publish(ex);
		}
	}
	
	/**
	 * If not all details are loaded (for all the given codes) 
	 * a new AirportsInfoLoader bahaviour is added and the next details 
	 * are going to be loaded.
	 */
	public int onEnd()
	{
	    super.onEnd();
		if (airLoader.morePages2Load())
		{
			((WrapperAgent)myAgent).addBusinessSubBehaviour(this);
		}
		
		return 0;
	}
}
