package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import org.apache.commons.httpclient.methods.PostMethod;
import org.w3c.dom.Document;
import semfly.exceptions.WebSourceLoadException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for AirportsInfoLoader behaviour.
 * 
 * This behaviour loads necessary details for some of the given codes. 
 * The important thing is that this behaviour does not load details for all 
 * the codes given at once! 
 * After loading some info from "details site" 
 * (http://www.airrouting.com/content/airportloc.html) for e.g. 5 codes 
 * it gives loaded resources (as DOM objects) to the AirportsInfoProcessor. 
 * And if not all details are loaded (for all the given codes) 
 * a new AirportsInfoLoader bahaviour is added and the next details are going to 
 * be loaded.
 */
public class AirportsInfoLoaderBL
{
    /**
     * URI address of a web page with airports details.
     */
	private String webSiteURI = 
		"http://www.airrouting.com/scripts/airportloc.asp";
	
	/**
	 * This const tells us how many pages with airport details is
	 * loaded befor giving them to the AirportInfoProcessor behaviour.
	 */
	private final int CodesStep = 10;
	
	/**
	 * Array with codes of airports for which info must be loaded.
	 */
	private String[] airportCodes = null;
	
	/**
	 * Current index of the airport code in airportCodes array.
	 */
	private int curCodeIdx = 0;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	/**
	 * Constructor. Sets airports codes for which all necessary details should 
	 * be loaded (airCodes param) and currend code index on zero.
	 * @param owner Agent - owner of this behaviour.
	 * @param airCodes Airports codes for which all necessary details should 
	 * be loaded.
	 */
	public AirportsInfoLoaderBL(Agent agent, String[] airCodes)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (airCodes == null) throw new 
	    	IllegalArgumentException("Airports codes (airCodes) can't be null");
	    
		airportCodes = airCodes;
		curCodeIdx = 0;
	}
	
	/**
	 * Loads details for airports codes (form airportCodes array).
	 * @return Loaded web pages with airports details.
	 * @throws WebSourceLoadException Loading exception.
	 */
	public Document[] loadResources() throws WebSourceLoadException
	{
		assert airportCodes != null;
		
		ArrayList docs = new ArrayList();
		int maxIdx = Math.min(curCodeIdx + CodesStep, airportCodes.length);
		for ( ; curCodeIdx < maxIdx; curCodeIdx++)
			docs.add(loadResource());
		
		return (Document[])docs.toArray(new Document[docs.size()]);
	}
	
	/**
	 * Loads a web page with airport details.
	 * @return Loaded web resource.
	 */
	protected Document loadResource() throws WebSourceLoadException
	{	
		assert airportCodes != null;
		assert curCodeIdx >= 0 && curCodeIdx < airportCodes.length;
		
		return loadAirportDetails(webSiteURI, airportCodes[curCodeIdx]);
	}
	
	/**
	 * Loads info about airports.
	 * @return Loaded web page with info about airports.
	 */
	protected Document loadAirportDetails(String pageURI, String airportCode)
		throws WebSourceLoadException
	{	
		WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		PostMethod httpMethod = new PostMethod(pageURI);
		httpMethod.addParameter("RequestLocation", airportCode);
		
		try
		{
			doc = webLoader.loadWebSiteHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadAirportDetails failed", ex, pageURI);
		}
		
		return doc;
	}
	
	/**
	 * Checks whether there are more pages with airports details to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curCodeIdx < airportCodes.length;
	}
	
	/**
	 * Sets web site URI. This method is used only for testing purposes.
	 * @param uri
	 */
	public void setWebSiteURI(String uri)
	{
	    webSiteURI = uri;
	}
}
