package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.List;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import semfly.dto.AirportInfo;
import semfly.exceptions.WebSourceParseException;
import semfly.onto.jena.Airport;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for AirportsInfoProcessor behaviour.
 * 
 * This class parses the given DOM objects with airports info by using Jaxen 
 * and then creates appropriate token to be send to the Coordinator
 * (by using WrapperSender bahaviour).
 */
public class AirportsInfoProcessorBL
{
    /**
     * Web pages with airports details.
     */
	private Document[] airports = null;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	/**
	 * Constructor.
	 * @param owner Agent - owner of this behaviour.
	 * @param airportsDetails Web pages with airports details.
	 */
	public AirportsInfoProcessorBL(Agent agent, Document[] airportsDetails)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (airportsDetails == null) throw new 
    		IllegalArgumentException("Airports details (airportsDetails) can't be null");
		
		airports = airportsDetails;
		myAgent = agent;
	}
	
	/**
	 * Processes pages with airports details. Extracts an importane info about
	 * airports and creates an appropriate Jena Model.
	 */
	public Model processResource() throws WebSourceParseException
	{
		Model model = ModelFactory.createDefaultModel();
		model.setNsPrefix(Airport.NSPrefix, Airport.NS);
		
		AirportInfo airportInfo;
		for (int i = 0; i < airports.length; i++)
		{
		    airportInfo = extractAirportInfo(airports[i]);
		    if (airportInfo != null)
		        addAirportInfo2Model(airportInfo, model);
		}
		
		return model;
	}
	
	/**
	 * Adds info about an airport to a Jena Model.
	 * @param airportInfo Info about an airport.
	 * @param airModel Existing Jena Model.
	 */
	private void addAirportInfo2Model(AirportInfo airportInfo, Model airModel)
	{
		assert airportInfo != null;
		assert airModel != null;
		
		Resource res = airModel.createResource(Airport.BaseResourceURI + 
								airportInfo.getIataCode());
		res.addProperty(Airport.iataCode, airportInfo.getIataCode());
		res.addProperty(Airport.icaoCode, airportInfo.getIcaoCode());
		res.addProperty(Airport.name, airportInfo.getName());
		res.addProperty(Airport.elevation, airportInfo.getElevation());
		res.addProperty(Airport.location, airportInfo.getLocation());
		//res.addProperty(Airport.latitude, airportInfo.getLatitude());
		//res.addProperty(Airport.longitude, airportInfo.getLongitude());
	}
	
	/**
	 * Extracts info about an airport from the web page given.
	 * @param airportDoc Web page with airport details.
	 * @return Object with an important info about an airport.
	 */
	private AirportInfo extractAirportInfo(Document airportDoc) 
		throws WebSourceParseException
	{
	    if (airportDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
		try
		{
		    checkPageFormat(airportDoc);
		    
			DOMXPath xpath = new DOMXPath("/HTML/BODY/TABLE[1]/TR[3]" +
					"/TD[1]/TABLE[1]/TR[1]/TD[1]/TABLE[2]/TR[1]/TD[1]" +
					"/text()");
		
			List nodes = xpath.selectNodes(airportDoc);
			if (nodes.size() == 0) return null;
			
			ArrayList details = new ArrayList();
			for (int i = 0; i < nodes.size(); i++)
			{
				String str = ((Node)nodes.get(i)).getNodeValue();
				str = str.replaceAll("\n", "");
				str = str.replaceAll("\t", "");
				str = str.trim();
				if (str.length() > 0)
				{
					details.add(str);
				}
			}
			
			return createAirportInfoObject(details);
		}
		catch (WebSourceParseException ex)
		{
		    throw ex;
		}
		catch (Exception ex)
		{
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractAirportInfo() failed", ex, 
			        airportDoc.getDocumentURI());
		}
	}
	
	/**
	 * Checks the page (airportDoc) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document airportDoc) 
		throws WebSourceParseException, JaxenException
	{
	    if (airportDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    
	    DOMXPath xpath = new DOMXPath("/HTML/BODY/TABLE[1]/TR[3]" +
				"/TD[1]/TABLE[1]/TR[1]/TD[1]/TABLE[2]");
	    
	    
	    
	    if (xpath.selectNodes(airportDoc).size() <= 0)
	    {
	        throw new WebSourceParseException(myAgent.getAID(),
	                "Format of the page is incorrect", 
	                airportDoc.getDocumentURI());
	    }
	}
	
	/**
	 * Creates object descrabing an airport form parameters in ArrayList.
	 * @param values Array with info about an airport.
	 * @return Object with an important info about an airport.
	 */
	private AirportInfo createAirportInfoObject(ArrayList values)
	{
	    assert values != null;
	    
	    int valuesSize = values.size();
	    if ( valuesSize == 0) return null;
	    
	    AirportInfo airportInfo = new AirportInfo();
		airportInfo.setName(values.get(0).toString());
		
		if (valuesSize > 1) airportInfo.setIataCode(values.get(1).toString());
		if (valuesSize > 2) airportInfo.setIcaoCode(values.get(2).toString());
		if (valuesSize > 3) airportInfo.setLocation(values.get(3).toString());
		//if (valuesSize > 4) airportInfo.setLatitude(0); - 4
		//if (valuesSize > 5) airportInfo.setLongitude(0); - 5
		if (valuesSize > 6) 
		    airportInfo.setElevation(Double.
		            parseDouble(values.get(6).toString()));
		
		return airportInfo;
	}
}
