package semfly.agents.behaviours;

import semfly.agents.CoordinatorAgent;
import semfly.exceptions.ExceptionManager;
import jade.content.AgentAction;
import jade.content.lang.Codec;
import jade.content.onto.OntologyException;
import jade.core.Agent;
import jade.core.behaviours.DataStore;
import jade.domain.FIPAAgentManagement.FailureException;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPAAgentManagement.RefuseException;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.proto.ContractNetResponder;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This behaviour belongs to Coordinator, is responsible for receiving 
 * 1. an IndexToken from a Wrapper Agent - Coordinator saves the received token
 * 		in its Tokens Priority Queue
 * 2. an RequestToken action from an Indexing Agent - Coordinator gets a token
 * 		with the highest priority form its Tokens Priority Queue and sends it to
 * 		the Indexing Agent
 * 
 * All the communication is done by using Contract-Net-Protocol. 
 */
public class CoordinatorReceiver extends ContractNetResponder
{
    /**
	 * Timeout for receiving messages in Contract-Net-Protocol.
	 */
    private final int ContractTimeout = 1000;
    
    /**
     * This class is responsible for "business logic" within this behaviour.
     */
    private CoordinatorReceiverBL receiverBL;
    
	/**
	 * Constructor.
	 * @param agent Agent - owner of this behaviour.
	 * @param msgTemplate Message tempale for received messages.
	 */
	public CoordinatorReceiver(Agent agent, MessageTemplate msgTemplate)
	{
		super(agent, msgTemplate);
		assert agent instanceof CoordinatorAgent;
		
		receiverBL = new CoordinatorReceiverBL(agent);
	}

	/**
	 * Constructor.
	 * @param agent Agent - owner of this behaviour.
	 * @param msgTemplate Message tempale for received messages.
	 * @param ds Agent's data store.
	 */
	public CoordinatorReceiver(Agent agent, MessageTemplate msgTemplate,
			DataStore ds)
	{
		super(agent, msgTemplate, ds);
		assert agent instanceof CoordinatorAgent;
		
		receiverBL = new CoordinatorReceiverBL(agent);
	}
	
	/**
	 * Prepares response for the received ACL message.
	 */
	protected ACLMessage prepareResponse(ACLMessage cfp) 
		throws NotUnderstoodException, RefuseException 
	{
	    ACLMessage response = cfp.createReply();
		try
		{
		    if (receiverBL.isActionForMe(receiverBL.extractAgentAction(cfp)))
		    {
		        response.setContent(((CoordinatorAgent)myAgent).
		                	getNumberOfTokens() + "");
		        response.setPerformative(ACLMessage.PROPOSE);
		   	}
			else
				response.setPerformative(ACLMessage.REFUSE);
		}
		catch (Codec.CodecException ex)
		{
			response.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			ExceptionManager.publish(ex);
		}
		catch (OntologyException ex)
		{
			response.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			ExceptionManager.publish(ex);
		}
		
		return response;
	}
	
	/**
	 * Prepares result notification for the requested action.
	 */
	protected ACLMessage prepareResultNotification(ACLMessage cfp, 
			ACLMessage propose, ACLMessage accept) throws FailureException 
	{
	    ACLMessage response = accept.createReply();
		
		try
		{
			AgentAction action = receiverBL.extractAgentAction(cfp);
			receiverBL.performAction(action, response);
			return response;
		}
		catch (Codec.CodecException ex)
		{
			ExceptionManager.publish(ex);
		}
		catch (OntologyException ex)
		{
			ExceptionManager.publish(ex);
		}
		catch (Exception ex)
		{
			ExceptionManager.publish(ex);
		}
		
		response.setPerformative(ACLMessage.FAILURE);
		return response;
	}
	
	/*
	protected void handleRejectProposal(ACLMessage cfp, ACLMessage propose, 
						ACLMessage reject) 
	{
	}
	*/
}
