package semfly.agents.behaviours;

import java.io.IOException;
import java.util.ArrayList;

import mo.agents.actions.FindMissingTokensAction;
import mo.dataprog.dataStructures.Token;
import mo.dataprog.utilities.PatternSearcher;
import mo.dataprog.utilities.WebPageReader;
import jade.content.AgentAction;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import semfly.agents.ITPQOwner;
import semfly.exceptions.AgentException;
import semfly.onto.jade.IndexToken;
import semfly.onto.jade.RequestAction;
import semfly.onto.jade.RequestToken;
import semfly.onto.jade.SaveAction;
import semfly.onto.jade.SaveInDB;
import semfly.onto.jade.SaveInTPQ;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for CoordinatorReceiver behaviour.
 */
public class CoordinatorReceiverBL
{
    /**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
    
	public CoordinatorReceiverBL(Agent agent)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    
	    myAgent = agent;
	}
	
	/**
	 * Performs requested action.
	 * @throws Exception 
	 * @throws IOException 
	 */
	public void performAction(AgentAction action, ACLMessage response)
		throws IOException, Exception
	{
	    if (action == null)
	        throw new IllegalArgumentException("AgentAction can't be null");
	    if (response == null) throw new 
	        IllegalArgumentException("Response (ACLMessage) can't be null");
	    
	    if (action instanceof SaveAction) 
	        performSaveAction((SaveAction)action);
	    else if (action instanceof RequestAction)
	        performRequestAction((RequestAction)action, response);
// Modification by Maciej Obojski 5-01-2007 -->
	    else if (action instanceof mo.agents.actions.FindMissingTokensAction)
	    	performFindMissingTokensAction((mo.agents.actions.FindMissingTokensAction)action, response);
// <-- Modification by Maciej Obojski 5-01-2007 	    
	    else
	        throw new AgentException(myAgent.getAID(), "Unknown Action");
	    
	    response.setPerformative(ACLMessage.INFORM);
	}	
	

	/**
	 * Extracts Agent Action from the message given.
	 * @param message
	 * @return Extracted action.
	 * @throws OntologyException
	 * @throws Codec.CodecException
	 */
	public AgentAction extractAgentAction(ACLMessage message) 
		throws OntologyException, Codec.CodecException
	{
	    if (message == null) throw new 
        	IllegalArgumentException("Message (ACL) can't be null");
		
		ContentElement ce = myAgent.getContentManager().extractContent(message);
		return (AgentAction)((Action)ce).getAction();
	}
	
    /**
	 * Checks if the owner of this behaviour is able to serve the action.
	 * @param action Action to perform.
	 * @return True - Owner of this behaviour is able to perform the action given,
	 * otherwise False.
	 */
	public boolean isActionForMe(AgentAction action)
	{
	    if (action == null)
	        throw new IllegalArgumentException("AgentAction can't be null");
	    
	    if (action instanceof SaveInTPQ)
	        return true;
	    else if (action instanceof RequestToken)
	        return ((ITPQOwner)myAgent).hasIndexTokens();
 // Modification by Maciej Obojski 5-01-2007 -->
	    else if (action instanceof mo.agents.actions.FindMissingTokensAction)
	    	return true;
// <-- Modification by Maciej Obojski 5-01-2007 	    
	    
	    return false;
	}
	
	/**
	 * Performs SaveAction - Coordinator saves a token in its TPQ.
	 * @param saveAction Action to be done.
	 * @throws AgentException
	 */
	private void performSaveAction(SaveAction saveAction) 
		throws AgentException
	{
	    assert saveAction != null;
	    
	    IndexToken token = saveAction.getIndexToken();
	    if (saveAction instanceof SaveInTPQ)
	        saveTokenInTPQ(token);
	    else
	        throw new AgentException(myAgent.getAID(), "Unknown Action");
	}
	
	/**
	 * Saves the token given in Tokens Priority Queue (TPQ) of the Coordinator. 
	 * This method is called when a CoordinatorAgent receives an IndexToken.
	 * @param token IndexToken to be saved in TPQ.
	 */
	private void saveTokenInTPQ(IndexToken token)
	{
	    if (token == null)
	        throw new IllegalArgumentException("IndexToken can't be null");
	    
	    assert myAgent instanceof ITPQOwner;
	    
	    ITPQOwner tpqOwner = (ITPQOwner)myAgent;
	    tpqOwner.addIndexToken2TPQ(token);
	}
	
	/**
	 * Performs RequestAction - Coordinator gives an IndexingAgent a token with
	 * the highest priority.
	 * @param requestAction
	 * @param response
	 * @throws AgentException
	 * @throws OntologyException
	 * @throws Codec.CodecException
	 */
	private void performRequestAction(RequestAction requestAction, 
	        ACLMessage response) throws AgentException, OntologyException,
			Codec.CodecException
	{
	    assert requestAction != null;
	    assert response != null;
	    
	    if (requestAction instanceof RequestToken)
	        requestIndexToken(response);
	    else
	        throw new AgentException(myAgent.getAID(), "Unknown Action");
	}
	
	/**
	 * Coordinator gives an IndexingAgent a token with
	 * the highest priority.
	 * @param response
	 * @throws OntologyException
	 * @throws Codec.CodecException
	 */
	private void requestIndexToken(ACLMessage response) throws OntologyException,
		Codec.CodecException
	{
	    assert response != null;
	    
	    SaveInDB saveAction = new SaveInDB();
	    saveAction.setIndexToken(((ITPQOwner)myAgent).poolIndexToken());
	    
	    Action contentAction = new Action();
  		contentAction.setAction(saveAction);
  		contentAction.setActor(myAgent.getAID());
  		
  		myAgent.getContentManager().fillContent(response, contentAction);
	}
	
	
	
/*****************************************
 ***  Methods added by Maciej Obojski  ***
 * @throws Exception 
 * @throws IOException 
 *****************************************/
	
// Modification by Maciej Obojski 5-01-2007 -->
	
	private void performFindMissingTokensAction(mo.agents.actions.FindMissingTokensAction action,
			ACLMessage response) throws IOException, Exception
	{
		assert action		!= null;
		assert response 	!= null;
		
		ArrayList<Token> patternToSearch = action.getPattern();
		
		PatternSearcher searcher = new PatternSearcher(action.getObjectToFindAddress(), patternToSearch);
		
		String found = searcher.getMax();
		
	//	myAgent.getContentManager().fillContent(response, =
		
		
	}
	
	
	
	
	
// <-- Modification by Maciej Obojski 5-01-2007
}
