package semfly.agents.behaviours;

import jade.wrapper.AgentController;
import jade.wrapper.ControllerException;
import jade.wrapper.PlatformController;

import java.util.ArrayList;
import java.util.Hashtable;

import semfly.agents.AgentInfo;
import semfly.agents.CoordinatorAgent;
import semfly.agents.WrapperAgentInfo;
import semfly.coordination.ScheduleReader;
import semfly.exceptions.ExceptionManager;
import semfly.globals.LongWrapper;

/**
 * @author szymek
 *
 * This class plays as a business logic for CoordinatorScheduler behaviour.
 * 
 * This class represents a scheduler of the Coordinator Agent.
 * This class is responsible for:
 * 1. starting Indexing Agents at CCS start-up
 * 2. starting Wrapper Agents at CCS start-up and starting them in appropriate
 * 		time (after appropriate intervales)
 */
public class CoordinatorSchedulerBL
{
    /**
     * Owner of this behaviour.
     */
    protected CoordinatorAgent myAgent;
    
    /**
     * Tells if at least one agent was started.
     */
	protected boolean acted;
	
	protected long currentTimestamp;
	
	protected PlatformController platform;
	protected WrapperAgentInfo[][] agentsSchedule;
	
	public CoordinatorSchedulerBL(CoordinatorAgent coordinatorAgent) 
	{
	    if (coordinatorAgent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    
	    myAgent = coordinatorAgent;
		setup();
	}

	protected void setup()
	{
		platform = null;
		agentsSchedule = null;
		
		acted = false;
		currentTimestamp = 0;
	}

	/**
	 * Starts IndexingAgents and all WrapperAgents. This method is called only
	 * once - at system start-up.
	 */
	public void actForTheFirstTime() throws Exception
	{
		startIndexingAgents();
		
		ScheduleReader scheduleReader = new ScheduleReader();
		agentsSchedule = scheduleReader.readWrappersScheduleSorted();
		WrapperAgentInfo[] eqSchedule;
		for (int i = 0; i < agentsSchedule.length; i++)
		{
			eqSchedule = agentsSchedule[i];
			startScheduledWrapperAgents(eqSchedule);
		}
		
		acted = true;
	}
	
	/**
	 * Waits for the proper time and starts Wrapper Agents who should be started 
	 * (based on passed time interval).
	 */
	public void actForTheNextTime() throws Exception
	{
		try 
		{
			WrapperAgentInfo[] eqSchedule;
			LongWrapper newTimestamp = new LongWrapper(0),
						sleepTime = new LongWrapper(0);
			
			eqSchedule = getNextScheduledAgents(newTimestamp, sleepTime);
			
			Thread.sleep(sleepTime.getValue());
			
			currentTimestamp = newTimestamp.getValue();
			startScheduledWrapperAgents(eqSchedule);
		}
		catch (InterruptedException ex)
		{
			ex.printStackTrace();
		}
	}
	
	/**
	 * Starts the WrapperAgents given.
	 * @param agents
	 */
	private void startScheduledWrapperAgents(WrapperAgentInfo[] agents)
	{
		assert agents != null;
		
		try
		{
			AgentController agent;
			for (int i = 0; i < agents.length; i++)
			{
				agent = platform.createNewAgent(agents[i].getNickname(), 
						agents[i].getClassName(), 
						agents[i].getParams());
				agent.start();
			}
		}
		catch (ControllerException ex)
		{
			ExceptionManager.publish(ex);
		}
	}
	
	/**
	 * Gets an array of Agents that should be started next time.
	 * 
	 * For each group of agents that should be started after the same interval,
	 * computes the closest time when they can be started.
	 * Agents that should be started next time are these whose closest time
	 * has the lowest value among closest times of the groups of agents.
	 * 
	 * @param newTimestamp Output param - tells the clotest time when agents 
	 * should be started.
	 * @return sleepTime Output param - tells how much time leaves before 
	 * starting agents.
	 */
	public WrapperAgentInfo[] getNextScheduledAgents(LongWrapper newTimestamp,
	        LongWrapper sleepTime)
	{
	    if (currentTimestamp < 0) throw new 
	    	IllegalArgumentException("Current timestamp can't be lower than 0");
	    if (agentsSchedule == null)
	        throw new IllegalArgumentException("Agents schedule can't be null");
	    if (agentsSchedule.length <= 0) throw new 
	    	IllegalArgumentException("Agents schedule array can't be empty");
	    
		Hashtable schedulesHT = new Hashtable();
		long tmpTimestamp = agentsSchedule[0][0].getTimestamp(),
				factor = currentTimestamp / tmpTimestamp + 1,
				minTimestamp = factor * tmpTimestamp;
		
		addToSchedulesHT(schedulesHT, minTimestamp, 0);
		for (int i = 1; i < agentsSchedule.length; i++)
		{
			tmpTimestamp = agentsSchedule[i][0].getTimestamp();
			
			factor = currentTimestamp / tmpTimestamp + 1; 
			tmpTimestamp = factor * tmpTimestamp;
			
			addToSchedulesHT(schedulesHT, tmpTimestamp, i);
			
			if (tmpTimestamp < minTimestamp) minTimestamp = tmpTimestamp;
		}
		
		newTimestamp.setValue(minTimestamp);
		sleepTime.setValue(newTimestamp.getValue() - currentTimestamp);
		return convertToArray((ArrayList)schedulesHT.get(new Long(minTimestamp)));
	}
	
	/**
	 * Helps in selection of Wrapper Agents that are to be started next time.
	 * @param schedulesHT
	 * @param timestamp
	 * @param arrIdx
	 */
	private void addToSchedulesHT(Hashtable schedulesHT, long timestamp, 
	        int arrIdx)
	{
		assert schedulesHT != null;
		assert timestamp >= 0;
		assert arrIdx >= 0;
		
		
		Long scheduleKey = new Long(timestamp);
		ArrayList tmpSchedules = (ArrayList)schedulesHT.get(scheduleKey);
		if (tmpSchedules == null) 
		{
			tmpSchedules = new ArrayList();
			tmpSchedules.add(agentsSchedule[arrIdx]);
		}
		else
		{
			tmpSchedules.add(agentsSchedule[arrIdx]);
		}
		schedulesHT.put(scheduleKey, tmpSchedules);
	}
	
	/**
	 * Converts an ArrayList with Wrappers' schedules info to an array with 
	 * objects of WrapperAgentInfo.
	 * @param schedules
	 * @return
	 */
	private WrapperAgentInfo[] convertToArray(ArrayList schedules)
	{
		assert schedules != null;
		
		ArrayList tmpSchedules = new ArrayList();
		WrapperAgentInfo[] tmpAgents;
		
		for (int i = 0; i < schedules.size(); i++)
		{
			tmpAgents = (WrapperAgentInfo[])schedules.get(i);
			
			for (int j = 0; j < tmpAgents.length; j++)
				tmpSchedules.add(tmpAgents[j]);
		}
		
		return (WrapperAgentInfo[])tmpSchedules.toArray(
						new WrapperAgentInfo[tmpSchedules.size()]);
	}
	
	/**
	 * Creates pool of indexing agents. Number of agents and name of their class
	 * is supplied be the ScheduleReader.readIndexingAgents() method.
	 * This method is called only once in actForTheFirstTime() method.
	 */
	private void startIndexingAgents() throws Exception
	{
		ScheduleReader scheduleReader = new ScheduleReader();
		AgentInfo[] agents = scheduleReader.readIndexingAgents();
		
		AgentController agent;
		for (int i = 0; i < agents.length; i++)
		{
			agent = platform.createNewAgent(agents[i].getNickname(), 
					agents[i].getClassName(), 
					agents[i].getParams());
			agent.start();
		}
	}
	
	/**
	 * Tells if at least one agent was started.
	 * True - was started, otherwise False.
	 */
	public boolean isActed()
	{
	    return acted;
	}
	
	/**
	 * Sets agents schedule array. This method is used only for testing purposes.
	 */
	public void setAgentsSchedule(WrapperAgentInfo[][] aSchedule)
	{
	    if (aSchedule == null)
	        throw new IllegalArgumentException("aSchedule can't be null");
	    
	    agentsSchedule = aSchedule;
	}
	
	/**
	 * Sets current timestamp. This method is used only for testing purposes.
	 */
	public void setCurrentTimestamp(long timestamp)
	{
	    currentTimestamp = timestamp;
	}
	
    public void setPlatform(PlatformController platformCtrl)
    {
        platform = platformCtrl;
    }
}
