package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelDocument;
import semfly.dto.HotelLocation;
import semfly.dto.ResourceDocument;
import semfly.dto.SheratonHotelCode;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class is responsible for loading Hilton Hotels Codes.
 */
public class HiltonHotelCodesLoaderBL
{
    private final String baseWebSiteURI = 
        "http://www.hilton.com/en/hi/hotels/search/index.jhtml?" +
        "searchType=city&city=City&_D%3AstateDropBox=+&_D%3AcountryDropBox=+" +
        "&countrySelectName=countryDropBox&_D%3Abrandzzz=+&brandzzz=false" +
        "&%2Fcom%2Fhilton%2Fsearch%2Fclient%2Fhandler%2FHandlerProxy.submit=true" +
        "&_D%3A%2Fcom%2Fhilton%2Fsearch%2Fclient%2Fhandler%2FHandlerProxy.submit=+" +
        "&_DARGS=%2Fen%2Fhi%2Fhotels%2Fsearch%2Fhome_tab_hotel_search.jhtml" +
        "&stateDropBox=&countryDropBox="; //NL
    
    private final String baseUSHotelsURI = 
        "http://www.hilton.com/en/hi/hotels/search/index.jhtml?" +
        "searchType=city&city=City&_D%3AstateDropBox=+&_D%3AcountryDropBox=+" +
        "&countrySelectName=countryDropBox&_D%3Abrandzzz=+&brandzzz=false" +
        "&%2Fcom%2Fhilton%2Fsearch%2Fclient%2Fhandler%2FHandlerProxy.submit=true" +
        "&_D%3A%2Fcom%2Fhilton%2Fsearch%2Fclient%2Fhandler%2FHandlerProxy.submit=+" +
        "&_DARGS=%2Fen%2Fhi%2Fhotels%2Fsearch%2Fhome_tab_hotel_search.jhtml" +
        "&countryDropBox=US&stateDropBox="; //NY
    
    private final String startPageWithHotelCodes = 
        "http://www.hilton.com/en/hi/hotels/search/newresults.jhtml?" +
		"searchType=city&sortCriteria=brand&showOnlyAvailable=false&" +
		"statusMsg=status_success.jhtml&showCompareError=false&" +
		"_D%3AselectedHotels=+&_D%3AselectedHotels=+&_D%3AselectedHotels=+&" +
		"_D%3AselectedHotels=+&_D%3AselectedHotels=+&_D%3AselectedHotels=+&" +
		"_D%3AselectedHotels=+&_D%3AselectedHotels=+&_D%3AselectedHotels=+&" +
		"_D%3AselectedHotels=+" +
		"&_DARGS=%2Fen%2Fhi%2Fhotels%2Fsearch%2Fnewresults.jhtml.10"+
		"&page=";

    private final String USStateProvinceQS = "stateProvince";
    
    private final String USACode = "US";
	
	/**
	 * This const tells us how many pages with hotel codes is
	 * loaded befor giving them to the HiltonInfoLoader behaviour.
	 */
	private final int CodesStep = 4;
	
	/**
	 * Array with hotels locations for which info must be loaded.
	 */
	private HotelLocation[] hotelsLoc;
	
	/**
	 * Current index of the hotel location in hotelsLoc array.
	 */
	private int curLocIdx = 0;
	
	private Agent myAgent;
   
    
    public HiltonHotelCodesLoaderBL(Agent agent, 
            	HotelLocation[] hotelsLocations)
    {
        myAgent = agent;
        hotelsLoc = hotelsLocations;
    }
    
    /**
     * Loads pages with Hilton hotels codes.
     * @return Loaded web pages.
     * @throws WebSourceLoadException
     */
	public HotelDocument[] loadResources() throws WebSourceLoadException
	{
	    ArrayList docs = new ArrayList();
		
	    int maxIdx = Math.min(curLocIdx + CodesStep, hotelsLoc.length);
		for ( ; curLocIdx < maxIdx; curLocIdx++)
		{
			docs.add(loadResource());
		}
		
		return (HotelDocument[])docs.toArray(new HotelDocument[docs.size()]);
	}
	
    /** (non-Javadoc)
     * @see semfly.agents.behaviours.ResourceLoader#loadResource()
     */
    protected ResourceDocument loadResource() throws WebSourceLoadException
    {
        assert hotelsLoc != null;
        assert curLocIdx >= 0 && curLocIdx < hotelsLoc.length;
        
        HotelLocation hotelLoc = hotelsLoc[curLocIdx];
        String uri;
    
        if (hotelLoc.isLocatedInUSState())
        {
            uri = baseUSHotelsURI + hotelLoc.getLocationCode();
        }
        else 
        {
            uri = baseWebSiteURI + hotelLoc.getLocationCode();
        }
        
        return new HotelDocument(loadHotelPage(uri));
    }
    
    protected Document[] loadHotelPage(String uri) 
		throws WebSourceLoadException
	{
	    WebSiteLoader webLoader = new WebSiteLoader();
		GetMethod httpMethod = new GetMethod(uri);
		
		try
		{
		    httpMethod.setFollowRedirects(true);
		    Document startPage = webLoader.loadWebSiteHTML(httpMethod);
		    int noPages = extractNumberOfHotelPages(startPage);
		    
		    if (noPages > 0)
		    {
		        Document[] pages = new Document[noPages];
		        pages[0] = startPage;
		        for (int idx = 1; idx < noPages; idx++)
		        {
		            httpMethod = 
			            new GetMethod(startPageWithHotelCodes + (idx+1));
			        pages[idx] = webLoader.loadWebSiteHTML(httpMethod);
			    }
		        return pages;
		    }
		    else 
	        {
		        return new Document[]{startPage};
	        }
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadHotelsPage() failed", ex, uri);
		}
	}
    
    /**
     * Returns number of additional pages (without start page) 
     * with hotels codes or -1 if there is no additional page. 
     * @param hotelDoc01
     * @return
     * @throws WebSourceParseException
     */
    private int extractNumberOfHotelPages(Document hotelDoc01) 
    	throws WebSourceParseException
    {
        //<a href="javascript:gotoPage(5)" class="pageline">5</a>
     
        try
        {
	        DOMXPath xpath = new DOMXPath("//A[@class='pageline']/text()"); 
	        
	        String code;
			List list = xpath.selectNodes(hotelDoc01);
			if (list != null && list.size() > 0) 
			{
			    Node node = (Node)list.get(list.size() - 1);
			    if (node != null)
			    {
			        return Integer.parseInt(node.getNodeValue());
			    }
			}
			
			return -1;
        }
        catch (JaxenException ex)
        {
            throw new WebSourceParseException(myAgent.getAID(), 
                    "Unable to extract number with hotels codes", ex);
        }   
    }
    
    /**
     * Extracts hotels codes from the array with web pages given.
     * @param hotels An array of web pages with hotels codes.
     * @return Extracted codes.
     * @throws JaxenException
     */
    public String[] extractHotelsCodes(HotelDocument[] hotels) 
    	throws WebSourceParseException
    { 
        String page = "";
        try
        {
	        ArrayList hotelsCodes = new ArrayList();
	        for (int i = 0; i < hotels.length; i++)
	        {
	            Document[] pages = hotels[i].getHotelCodesPageColl();
	            if (pages == null) 
	                throw new WebSourceParseException(myAgent.getAID(), 
	                        	"unknown",
                    			"Unable to extract number with hotels codes");
	            
	            for (int j = 0; j < pages.length; j++)
	            {
	                //TODO: check page format - maybe in future ;-)
	                extractHotelsCodes(pages[j], hotelsCodes);
	            }
	        }
	        
	        return (String[])hotelsCodes.toArray(
	                	new String[hotelsCodes.size()]);
	    }
        /*
        catch (WebSourceParseException ex) 
        {
            throw ex;
        }
        */
        catch (JaxenException ex)
        {
            ex.printStackTrace();
            throw new WebSourceParseException(myAgent.getAID(), 
                    "extractHotelsCodes() failed", ex, "Hilton");
        }
    }
    
    protected void extractHotelsCodes(Document hotels, ArrayList hotelCodes)
    	throws JaxenException
    {
        assert hotels != null;
        assert hotelCodes != null;
        
		DOMXPath xpath = new DOMXPath("//TD[@class='hotelsearchHotelName']/A");
		
		String hotelLink, code;
		List nodes = xpath.selectNodes(hotels);
		Iterator nodesIter = nodes.iterator();
		Node n;
		int idx;
		while (nodesIter.hasNext())
		{
		    n = (Node)nodesIter.next();
		    
		    hotelLink = n.getAttributes().getNamedItem("href").getNodeValue();
		    if (hotelLink.indexOf("hilton.com") > 0 )
		    {
		        idx = hotelLink.lastIndexOf("ctyhocn=");
		        code = hotelLink.substring(idx + 8);
		        hotelCodes.add(code);
		    }
		}
    }
    
    private SheratonHotelCode createHotelCode(String hotelLink)
    {
        SheratonHotelCode hotelCode = new SheratonHotelCode();
        
        int idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setPropertyID(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
        
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setBrand(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
        
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setChainCode(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
       
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setAffilationCode(hotelLink.substring(0, idx));
        
        return hotelCode;
    }
    
    /**
	 * Checks the page (hotels) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document hotels) 
		throws WebSourceParseException, JaxenException
	{
	    if (hotels == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    DOMXPath xpath = new DOMXPath("//DIV[@class='propertyContainer']");
	    
	    if (xpath.selectNodes(hotels).size() <= 0)
	    {
	        xpath = new DOMXPath("//TD[@class='error']");

			if (xpath.selectNodes(hotels).size() <= 0)
			{
			    throw new WebSourceParseException(myAgent.getAID(),
	                	hotels.getDocumentURI(),
		                "Format of the page is incorrect");
			}
		}
	}
	
	/**
	 * Checks whether there are more pages with hotels codes to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curLocIdx < hotelsLoc.length;
	}
}
