package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import org.apache.commons.httpclient.methods.GetMethod;
import org.w3c.dom.Document;
import semfly.dto.HotelDocument;
import semfly.dto.ResourceDocument;
import semfly.exceptions.WebSourceLoadException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for HiltonInfoLoader behaviour.
 * 
 * This class is responsible loading info about Hilton hotels for the hotels
 * codes given from www.hilton.com web site.
 */
public class HiltonInfoLoaderBL
{
    /**
     * The array of hotel's codes that details are to be loaded.
     */
    private String[] hotelCodes;
    
    /**
     * part of URI addresses of a web page with Hilton hotels details.
     */
	private String webSiteURIBegining = "http://www.hilton.com/en/hi/hotels/";
	private String amenitiesURIEnding =
			"accommodations_amenities_popup.jhtml?roomType=STD&ctyhocn=";
	private String servicesURIEnding = "services.jhtml?ctyhocn=";
	
	
	/**
	 * This const tells us how many pages with hotels details is
	 * loaded befor giving them to the SheratonInfoProcessor behaviour.
	 */
	private final int CodesStep = 1;
	
	/**
	 * Current index of the hotel code in hotelCodes array.
	 */
	private int curCodeIdx = 0;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	/**
	 * Constructor. Sets hotels codes for which all necessary details should 
	 * be loaded (hotelsCodes param) and currend code index on zero.
	 * @param owner Agent - owner of this behaviour.
	 * @param hCodes Hilton hotels codes for which all necessary 
	 * details should be loaded.
	 */
	public HiltonInfoLoaderBL(Agent agent, String[] hCodes)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (hCodes == null) throw new 
	    	IllegalArgumentException("Hotels codes (hCodes) can't be null");
	    
	    myAgent = agent;
		hotelCodes = hCodes;
		curCodeIdx = 0;
	}
	
	/**
	 * Loads details for hotels codes (form hotelCodes array).
	 * @return Loaded web pages with hotels details.
	 * @throws WebSourceLoadException Loading exception.
	 */
	public HotelDocument[] loadResources() throws WebSourceLoadException
	{
		assert hotelCodes != null;
		
		HotelDocument hotelDoc;
		ArrayList docs = new ArrayList();
		int maxIdx = Math.min(curCodeIdx + CodesStep, hotelCodes.length);
		for ( ; curCodeIdx < maxIdx; curCodeIdx++)
		{
		    hotelDoc = (HotelDocument)loadResource();
		    hotelDoc.setHotelCode(hotelCodes[curCodeIdx]);
		    docs.add(hotelDoc);
		}
		
		 if (docs.size() <= 0) throw new
	        WebSourceLoadException(myAgent.getAID(), "unknown URI", 
	        						"loadResources(): empty array of web pages");
		
		return (HotelDocument[])docs.toArray(new HotelDocument[docs.size()]);
	}
	
	/**
	 * Loads a web page with hotel details.
	 * @return Loaded web resource.
	 */
	protected ResourceDocument loadResource() throws WebSourceLoadException
	{	
	    if (hotelCodes == null) throw new 
	    	WebSourceLoadException(myAgent.getAID(), 
	    	        "unknown URI", "hotelCodes == null");
	    
	    if (curCodeIdx < 0 || curCodeIdx >= hotelCodes.length) throw new
	        WebSourceLoadException(myAgent.getAID(), "unknown URI", 
	        						"out of hotelCodes array");
	    
		String amenitiesPage = 
		    webSiteURIBegining + amenitiesURIEnding + hotelCodes[curCodeIdx];
		String servicesPage = 
		   webSiteURIBegining + servicesURIEnding + hotelCodes[curCodeIdx];
		
		
	    HotelDocument hotelDoc = new HotelDocument();
	    hotelDoc.setAmenitiesPage(loadHotelDetails(amenitiesPage));
	    hotelDoc.setServicesPage(loadHotelDetails(servicesPage));
		return hotelDoc;
	}
	
	/**
	 * Loads info about Marriott hotels.
	 * @return Loaded web page with info about airports.
	 */
	protected Document loadHotelDetails(String hotelPage) 
		throws WebSourceLoadException
	{	
		WebSiteLoader webLoader = new WebSiteLoader();
		
		GetMethod httpMethod = new GetMethod(hotelPage);
		
		try
		{
			return webLoader.loadWebSiteXHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadHotelDetails failed", ex, hotelPage);
		}
	}
	
	/**
	 * Checks whether there are more pages with hotels details to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curCodeIdx < hotelCodes.length;
	}
	
	/**
	 * Sets web site beginning URI. This method is used only for testing purposes.
	 */
	public void setWebSiteURIBegining(String uri)
	{
	    webSiteURIBegining = uri;
	}
}
