package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelDocument;
import semfly.dto.HotelInfo;
import semfly.exceptions.WebSourceParseException;
import semfly.onto.jena.Hotel;
import semfly.onto.jena.Location;
import semfly.onto.jena.Vcard;
import semfly.textmining.HiltonTextMiner;
import semfly.textmining.TextDescription;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for HiltonInfoProcessor behaviour.
 *
 * This behaviour parses the DOM objects given with Hilton hotels info 
 * by using Jaxen and then creates an appropriate token to be send to 
 * the Coordinator (by using WrapperSender bahaviour).
 */
public class HiltonInfoProcessorBL
{
    /**
     * Web pages with Hilton hotels details.
     */
	private HotelDocument[] hotels = null;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	private Model hotelModel;
	
	/**
	 * Constructor.
	 * @param owner Agent - owner of this behaviour.
	 * @param hiltonDetails Web pages with hotels details.
	 */
	public HiltonInfoProcessorBL(Agent agent, HotelDocument[] hiltonDetails)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (hiltonDetails == null) throw new IllegalArgumentException(
	            "Hilton hotels details (hiltonDetails) can't be null");
	   
		hotels = hiltonDetails;
		myAgent = agent;
	}
	
	/**
	 * Processes pages with hotels details. Extracts an important info about
	 * hotels and creates an appropriate Jena Model.
	 */
	public Model processResource() throws WebSourceParseException
	{
	    hotelModel = ModelFactory.createDefaultModel();
		
		HotelInfo hotelInfo;
		Resource hotel;
		for (int i = 0; i < hotels.length; i++)
		{
		    hotelInfo = extractHotelBaseInfo(hotels[i].getAmenitiesPage());
		    hotelInfo.setHotelCode(hotels[i].getHotelCode().toString());
		    
		    hotel = addHotelInfo2Model(hotelInfo, hotelModel);
		    extractHotelInfo(hotels[i], hotel);
		}
		
		return hotelModel;
	}
	
	/**
	 * Adds info about a hotel to a Jena Model.
	 * @param hotelInfo Info about a hotel.
	 * @param hotelModel Existing Jena Model.
	 */
	private Resource addHotelInfo2Model(HotelInfo hotelInfo, Model hotelModel)
	{
	    String hotelURI = Hotel.BaseResourceURI + "Hilton/" + 
	    					hotelInfo.getHotelCode();
	    
		Resource hotel = hotelModel.createResource(hotelURI);
		hotel.addProperty(Hotel.id, "Hilton/" + hotelInfo.getHotelCode());
		
		//address
		Resource address = hotelModel.createResource(hotelURI + "/Address");
		
		String value;
		value = hotelInfo.getStreet();
		if (value != null && value.length() > 0)
		    address.addProperty(Vcard.Street, hotelInfo.getStreet());
		
		value = hotelInfo.getLocality();
		if (value != null && value.length() > 0)
		    address.addProperty(Vcard.Locality, hotelInfo.getLocality());
		
		value = hotelInfo.getPcode();
		if (value != null && value.length() > 0)
		    address.addProperty(Vcard.Pcode, hotelInfo.getPcode());
		
		value = hotelInfo.getCountry();
		if (value != null && value.length() > 0)
			address.addProperty(Vcard.Country, hotelInfo.getCountry());
		
		value = hotelInfo.getPhone();
		if (value != null && value.length() > 0)
		    hotel.addProperty(Location.phone, hotelInfo.getPhone());
		
		hotel.addProperty(Location.address, address);
		return hotel;
	}
	
	/**
	 * Extracts info about a hotel from the web page given.
	 * @param hotelDoc Web page with hotel details.
	 * @return Object with an important info about a hotel.
	 */
	private HotelInfo extractHotelBaseInfo(Document hotelDoc) 
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	   
	    HotelInfo hotelInfo = new HotelInfo();
	    try
	    {
			extractHotelAddress(hotelDoc, hotelInfo);
			
			return hotelInfo;
	    }
	    catch (Exception ex)
	    {
	        ex.printStackTrace();
	    }
	    
	    return hotelInfo;
	}
	
	/**
	 * Extracts info about a hotel.
	 * @param hotelDoc Web pages with hotel details.
	 * @return Object with an important info about a hotel.
	 */
	private void extractHotelInfo(HotelDocument hotelDoc, Resource hotelRes) 
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Documents to parse can't be null");
	
	    extractRoomAmenityInfo(hotelDoc.getAmenitiesPage(), hotelRes);
	    extractSiteAmenityInfo(hotelDoc.getServicesPage(), hotelRes);
		extractCheckInCheckOutInfo(hotelDoc.getServicesPage(), hotelRes);
		extractPetsPolicyInfo(hotelDoc.getServicesPage(), hotelRes);
		extractBusinessServicesInfo(hotelDoc.getServicesPage(), hotelRes);
		extractRecreationInfo(hotelDoc.getServicesPage(), hotelRes);
	}
	
	private void extractRoomAmenityInfo(Document hotelDoc, 
	        		Resource hotelRes) throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
	    try
		{
	        DOMXPath xpath = 
		        new DOMXPath("//TD[@class='defaultText']/text()");
			
			List nodes = xpath.selectNodes(hotelDoc);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			Node node;
			ArrayList roomInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    roomInfo.add(nodeValue);
			}
			
			parseRoomAmenityInfo(
		            (String[])roomInfo.toArray(new String[roomInfo.size()]), 
		            	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseRoomAmenityInfo(String[] amenities, Resource hotelRes)
	{
	    HiltonTextMiner txtMiner = HiltonTextMiner.getInstance();
	    TextDescription txtDesc;
	    for (int i = 0; i < amenities.length; i++)
	    {
	        txtDesc = txtMiner.getAmenityDescription(amenities[i]);
	        if (txtDesc != null)
	        {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
	        }
	    }
	}
	
	private void extractHotelAddress(Document hotelPage, HotelInfo hotelInfo)
		throws WebSourceParseException
	{
	    if (hotelPage == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
	    try
		{
	        DOMXPath xpath = 
		        new DOMXPath("//SPAN[@class='hotelName']/text()");
	        Node node = (Node)xpath.selectSingleNode(hotelPage);
			if (node == null) return;
			
			hotelInfo.setName(node.getNodeValue());
	        
		    xpath = new DOMXPath("//SPAN[@class='hotelAddress']/text()[1]");
			node = (Node)xpath.selectSingleNode(hotelPage);
			if (node == null) return;
			parseHotelAddress(node.getNodeValue(), hotelInfo);
			
			xpath = new DOMXPath("//SPAN[@class='hotelAddress']/text()[2]");
			node = (Node)xpath.selectSingleNode(hotelPage);
			if (node == null) return;
			parseHotelPhoneFax(node.getNodeValue(), hotelInfo);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelPage.getDocumentURI());
		}
	}
	
	private void parseHotelAddress(String nodeValue, HotelInfo hotelInfo)
	{
	    nodeValue = nodeValue.replaceAll("\n", "");
	    nodeValue = nodeValue.replaceAll("\t", "");
	    nodeValue = nodeValue.trim();
	    
	    int idx01, idx02, idx03;
	    
	    //extract street
	    idx01 = nodeValue.indexOf(',');
	    if (idx01 > 0)
	    {
		    String street = nodeValue.substring(0, idx01);
		    hotelInfo.setStreet(street);
	    }
	    
	    //extract locality
	    idx02 = nodeValue.lastIndexOf(',');
	    if (idx01 + 2 > idx02 && idx02 > 0)
	    {
		    String locality = nodeValue.substring(idx01 + 2, idx02);
		    hotelInfo.setLocality(locality);
	    }
	    
	    //extract country and postal code
	    String delimeter = nodeValue.substring(idx01 + 1, idx01 + 2);
	    idx03 = nodeValue.lastIndexOf(delimeter);
	    if (idx03 > idx02 + 2 && idx02 > 0)
	    {
		    String country = nodeValue.substring(idx02 + 2, idx03);
		    hotelInfo.setCountry(country);
		    String pcode = nodeValue.substring(idx03 + 1);
		    hotelInfo.setPcode(pcode);
	    }
	}
	
	private void parseHotelPhoneFax(String nodeValue, HotelInfo hotelInfo)
	{
	    nodeValue = nodeValue.replaceAll("\n", "");
	    nodeValue = nodeValue.replaceAll("\t", "");
	    nodeValue = nodeValue.trim();
	    int idx;
	    
	    //extract phone
	    idx = nodeValue.indexOf(':');
	    if (idx > 0)
	    {
		    String delimeter = nodeValue.substring(idx + 1, idx + 2);
		    if (delimeter != null && delimeter.length() > 0)
		    {
		        String phone = nodeValue.substring(idx + 2);
		        idx = phone.indexOf(delimeter);
		        phone = phone.substring(0, idx);
		        hotelInfo.setPhone(phone);
		    }
	    }
	    
	    //extract fax
	    idx = nodeValue.lastIndexOf(':');
	    if (idx > 0)
	    {
		    String delimeter = nodeValue.substring(idx + 1, idx + 2);
		    if (delimeter != null && delimeter.length() > 0)
		    {
		        String fax = nodeValue.substring(idx + 2);
		        idx = fax.indexOf(delimeter);
		        fax = fax.substring(0, idx);
		        hotelInfo.setFax(fax);
		    }
	    }
	}
	
	private void extractCheckInCheckOutInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
	        //<td class="defaultText"><b>Check-in/Check-out</b></td>
		    DOMXPath xpath = 
		        new DOMXPath("//TD[@class='defaultText']/" +
		        		"B[text() = 'Check-in/Check-out']");
		
		    Node node = (Node)xpath.selectSingleNode(hotelDoc);
		    if (node == null) return;
		    
		    StringBuffer checkInOut = new StringBuffer();
		    
		    //get check-in info
		    xpath = new DOMXPath("ancestor::TR/following-sibling::TR" +
		    		"[position()=2]/TD[@class='defaultText']/text()");
		    
		    Node checkNode = (Node)xpath.selectSingleNode(node);
		    if (checkNode != null)
		        checkInOut.append(checkNode.getNodeValue().trim());
		    
		    //get check-out info
		    xpath = new DOMXPath("ancestor::TR/following-sibling::TR" +
		    		"[position()=3]/TD[@class='defaultText']/text()");
		    
		    checkNode = (Node)xpath.selectSingleNode(node);
		    if (checkNode != null)
		    {
		        checkInOut.append("; ");
		        checkInOut.append(checkNode.getNodeValue().trim());
		    }
		    
			Resource checkRes = hotelModel.createResource(
		            hotelRes.getURI() + "/CheckIn-CheckOut");
		    checkRes.addProperty(Hotel.additionalDetail, checkInOut.toString());
		    hotelRes.addProperty(Hotel.additionalDetail, checkRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void extractPetsPolicyInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
	        //<td width="150" class="defaultText"><b>Pets:
		    DOMXPath xpath = 
		        new DOMXPath("//TD[@class='defaultText']/" +
		        		"B[text() = 'Pets:']");
		
		    Node node = (Node)xpath.selectSingleNode(hotelDoc);
		    if (node == null) return;
		    
		    //get pets policy: <td class="defaultText">Allowed</td>
		    xpath = new DOMXPath("ancestor::TD/following-sibling::TD" +
		    		"[position()=1 and @class='defaultText']/text()");
		    
		    node = (Node)xpath.selectSingleNode(node);
		    if (node != null)
		    {
		        HiltonTextMiner txtMiner = HiltonTextMiner.getInstance();
			    TextDescription txtDesc;
			    txtDesc = txtMiner.getPetsPolicyDescription(
			            	node.getNodeValue().trim());
			    
			    if (txtDesc != null)
			    {
			        hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
			    }
		    }
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void extractSiteAmenityInfo(Document hotelDoc, 
    		Resource hotelRes) throws WebSourceParseException
	{
		if (hotelDoc == null)
		    throw new IllegalArgumentException("Document to parse can't be null");
		
		try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//TD[@class='headerBrand']/" +
		        		"ancestor::TR[position()=1]/" +
		        		"following-sibling::TR[position()=6]");
			
		    Node node = (Node)xpath.selectSingleNode(hotelDoc);
		    if (node == null) return;
		    
		    xpath = 
		        new DOMXPath("descendant::TD[@class='defaultText']/text()");
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList siteAmenity = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    siteAmenity.add(nodeValue);
			}
			
			parseSiteAmenityInfo(
			        (String[])siteAmenity.toArray(
			                new String[siteAmenity.size()]), hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractSiteAmenityInfo() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseSiteAmenityInfo(String[] amenities, Resource hotelRes)
	{
		HiltonTextMiner txtMiner = HiltonTextMiner.getInstance();
		TextDescription txtDesc;
		for (int i = 0; i < amenities.length; i++)
		{
		    txtDesc = txtMiner.getSiteAmenityDescription(amenities[i]);
		    if (txtDesc != null)
		    {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
		    }
		}
	}
	
	private void extractBusinessServicesInfo(Document hotelDoc, 
    		Resource hotelRes) throws WebSourceParseException
	{
		if (hotelDoc == null)
		    throw new IllegalArgumentException("Document to parse can't be null");
		
		try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//TD[@class='headerBrand']");
			
		    List nodes = xpath.selectNodes(hotelDoc);
		    if (nodes.size() < 2) return;
		    
		    Node node = (Node)nodes.get(1);
		    
		    xpath = 
		        new DOMXPath("ancestor::TR[position()=1]/" +
		        		"following-sibling::TR[position()=6]/" +
		        		"descendant::TD[@class='defaultText']/text()");
			
		    nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList businessSrvInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    businessSrvInfo.add(nodeValue);
			}
			
			parseBusinessServicesInfo(
		            (String[])businessSrvInfo.toArray(
		                    new String[businessSrvInfo.size()]), 
		                    	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractBusinessServicesInfo() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseBusinessServicesInfo(String[] services, Resource hotelRes)
	{
		HiltonTextMiner txtMiner = HiltonTextMiner.getInstance();
		TextDescription txtDesc;
		for (int i = 0; i < services.length; i++)
		{
		    txtDesc = txtMiner.getBusinessServiceDescription(services[i]);
		    if (txtDesc != null)
		    {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
		    }
		}
	}
	
	private void extractRecreationInfo(Document hotelDoc, 
    		Resource hotelRes) throws WebSourceParseException
	{
		if (hotelDoc == null)
		    throw new IllegalArgumentException("Document to parse can't be null");
		
		try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//TD[@class='headerBrand']");
			
		    List nodes = xpath.selectNodes(hotelDoc);
		    if (nodes.size() < 3) return;
		    
		    Node node = (Node)nodes.get(2);
		    
		    xpath = 
		        new DOMXPath("ancestor::TR[position()=1]/" +
		        		"following-sibling::TR[position()=6]/" +
		        		"descendant::TD[@class='defaultText']/text()");
			
		    nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList recreationInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    recreationInfo.add(nodeValue);
			}
			
			parseRecreationInfo(
		            (String[])recreationInfo.toArray(
		                    new String[recreationInfo.size()]), 
		                    	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractRecreationInfo() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseRecreationInfo(String[] services, Resource hotelRes)
	{
		HiltonTextMiner txtMiner = HiltonTextMiner.getInstance();
		TextDescription txtDesc;
		for (int i = 0; i < services.length; i++)
		{
		    txtDesc = txtMiner.getRecreationDescription(services[i]);
		    if (txtDesc != null)
		    {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
		    }
		}
	}
}
