package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelLocation;
import semfly.dto.WebPageDocument;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for HiltonLocationsLoader behaviour.
 * 
 * This class represents a Behaviour that is to load locations (their codes) 
 * of all Hilton Hotels.
 */
public class HiltonLocationsLoaderBL
{
    private String pageURI = "http://www.hilton.com/en/hi/index.jhtml";

	private Agent myAgent;
	
	public HiltonLocationsLoaderBL(Agent owner)
	{
	    myAgent = owner;
	}
	
	/**
	 * Loads locations of Hilton Hotels.
	 * @return
	 * @throws WebSourceLoadException
	 */
	public WebPageDocument loadLocations() throws WebSourceLoadException
	{
	    WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		GetMethod httpMethod = new GetMethod(pageURI);
		try
		{
			doc = webLoader.loadWebSiteXHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadLocations() failed", ex, pageURI);
		}
		
		return new WebPageDocument(doc);
	}
	
	/**
	 * Extracts locations codes from a web page with hotels locations.
	 * @param locations
	 * @return
	 * @throws WebSourceParseException
	 */
	public HotelLocation[] extractLocationCodes(Document locations) 
		throws WebSourceParseException
	{
	    if (locations == null)
	        throw new IllegalArgumentException(
	                "Page with hotels locations can't be null");
	    
	    try
	    {
	        checkPageFormat(locations);
	        
	        ArrayList hotels = new ArrayList();
	        extractCountryLocations(locations, hotels);
	        extractUSStateLocations(locations, hotels);
	        
	        return (HotelLocation[])hotels.toArray(
	                	new HotelLocation[hotels.size()]);
	    }
	    catch (WebSourceParseException ex)
		{
		    throw ex;
		}
	    catch (Exception ex)
	    {
	        throw new WebSourceParseException(myAgent.getAID(), 
			        "extractLocationCodes() failed", ex, pageURI);
	    }
	}
	
	/**
	 * Extracts locations codes for countries from a web page 
	 * with hotels locations.
	 * @param locations
	 * @param hotels
	 * @throws JaxenException
	 */
	private void extractCountryLocations(Document locations, ArrayList hotels)
		throws JaxenException
	{
	    assert locations != null;
	    assert hotels != null;
	    
		DOMXPath xpath = new DOMXPath("//SELECT[@name='countryDropBox']/" +
		    		"OPTION[@value != 'US']"); 
		
		String code;
		List nodes = xpath.selectNodes(locations);
		Iterator nodesIter = nodes.iterator();
		while (nodesIter.hasNext())
		{
		    code = extractLocationCode((Node)nodesIter.next());
		    if ( !locationCodeExists(code, hotels, true) )
		        hotels.add(new HotelLocation(code, HotelLocation.Country));
		}
	}
	
	/**
	 * Extracts locations codes for states in the USA from a web page 
	 * with hotels locations.
	 * @param locations
	 * @param hotels
	 * @throws JaxenException
	 */
	private void extractUSStateLocations(Document locations, ArrayList hotels)
		throws JaxenException
	{
	    assert locations != null;
	    assert hotels != null;
	   
		DOMXPath xpath = new DOMXPath("//SELECT[@name='stateDropBox']" +
				"/OPTION[@value != '']");
		
		String code;
		int noEmptyCode = 0;
		List nodes = xpath.selectNodes(locations);
		Iterator nodesIter = nodes.iterator();
		while (nodesIter.hasNext())
		{
		    code = extractLocationCode((Node)nodesIter.next());
		    if (code.equals("AB")) break;
		    if ( !locationCodeExists(code, hotels, false) )
		        hotels.add(new HotelLocation(code, HotelLocation.USState));
		}
	}
	
	private boolean locationCodeExists(String code, ArrayList hotels, 
	        			boolean checkCountry)
	{
	    String exCode;
	    HotelLocation location;
	    for (int i = 0; i < hotels.size(); i++)
	    {
	        location = (HotelLocation)hotels.get(i);
	           
	        if (location.getLocationType() == HotelLocation.Country && 
	                checkCountry)
	        {
	            if (location.getLocationCode().equals(code)) return true;
	        }
	        else if (location.getLocationType() == HotelLocation.USState && 
	                !checkCountry)
	        {
	            if (location.getLocationCode().equals(code)) return true;
	        }
	    }
	    
	    return false;
	}
	
	/**
	 * Extracts location code from the node given.
	 * @param node
	 * @return
	 */
	private String extractLocationCode(Node node)
	{
	    assert node != null;
	    
	    return node.getAttributes().getNamedItem("value").getNodeValue();
	}
	
	/**
	 * Checks the page (locations) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document locations) 
		throws WebSourceParseException, JaxenException
	{
	    if (locations == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    DOMXPath xpath = new DOMXPath("//SELECT[@name='countryDropBox']/" +
	    		"OPTION[@value != 'US']");  
	    
	    if (xpath.selectNodes(locations).size() <= 0)
	    {
	        xpath = new DOMXPath("//SELECT[@name='stateDropBox']" +
					"/OPTION[@value != '']");
	        
	        if (xpath.selectNodes(locations).size() <= 0)
		    {
	            throw new WebSourceParseException(myAgent.getAID(),
		                "Format of the page is incorrect", 
		                locations.getDocumentURI());
		    }
	    }
	}
	
	/**
	 * This method is used only for testing purposes.
	 * @param uri
	 */
	public void setPageURI(String uri)
	{
	    pageURI = uri;
	}
}
