package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelLocation;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class is responsible for loading Marriot Hotels Codes.
 */
public class MarriottHotelCodesLoaderBL
{
    private String baseWebSiteURI = 
		"http://marriott.com/search/hotelDirectory.mi?";//country=AU&bc=";
    
    private final String CountryQS = "country",
    						USStateQS = "state",
    						BC = "bc";
	
	/**
	 * This const tells us how many pages with hotel codes is
	 * loaded befor giving them to the MarriottInfoLoader behaviour.
	 */
	private final int CodesStep = 4;
	
	/**
	 * Array with hotels locations for which info must be loaded.
	 */
	private HotelLocation[] hotelsLoc;
	
	/**
	 * Current index of the hotel location in hotelsLoc array.
	 */
	private int curLocIdx = 0;
	
	private Agent myAgent;
   
    
    public MarriottHotelCodesLoaderBL(Agent agent, 
            HotelLocation[] hotelsLocations)
    {
        myAgent = agent;
        setHotelsLoc(hotelsLocations);
    }
    
    /**
     * Loads pages with Marriott hotels codes.
     * @return Loaded web pages.
     * @throws WebSourceLoadException
     */
	public Document[] loadResources() throws WebSourceLoadException
	{
	    ArrayList docs = new ArrayList();
		int maxIdx = Math.min(curLocIdx + CodesStep, hotelsLoc.length);
		for ( ; curLocIdx < maxIdx; curLocIdx++)
			docs.add(loadResource());
		
		return (Document[])docs.toArray(new Document[docs.size()]);
	}
	
    /** (non-Javadoc)
     * @see semfly.agents.behaviours.ResourceLoader#loadResource()
     */
    protected Document loadResource() throws WebSourceLoadException
    {
        assert hotelsLoc != null;
        assert curLocIdx >= 0 && curLocIdx < hotelsLoc.length;
        
        HotelLocation hotelLoc = hotelsLoc[curLocIdx];
        String uri = baseWebSiteURI;
        if (hotelLoc.isLocatedInCountry()) uri += CountryQS;
        else if (hotelLoc.isLocatedInUSState()) uri += USStateQS;
        
        uri += "=" + hotelLoc.getLocationCode();
        uri += "&" + BC + "=";
        
        return loadHotelsPage(uri);
    }
    
    protected Document loadHotelsPage(String pageURI) 
    	throws WebSourceLoadException
    {
        assert pageURI != null;
        
        WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		GetMethod httpMethod = new GetMethod(pageURI);
		try
		{
			doc = webLoader.loadWebSiteXHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadHotelsPage() failed", ex, pageURI);
		}
		
		return doc;
    }
    
    /**
     * Extracts hotels codes from the array with web pages given.
     * @param hotels An array of web pages with hotels codes.
     * @return Extracted codes.
     * @throws JaxenException
     */
    public String[] extractHotelsCodes(Document[] hotels) throws 
    	WebSourceParseException
    {
        try
        {
	        ArrayList hotelsCodes = new ArrayList();
	        
	        for (int i = 0; i < hotels.length; i++)
	        {
	            checkPageFormat(hotels[i]);
	            extractHotelsCodes(hotels[i], hotelsCodes);
	        }
	        
	        return (String[])hotelsCodes.toArray(new String[hotelsCodes.size()]);
	    }
        catch (JaxenException ex)
        {
            throw new WebSourceParseException(myAgent.getAID(), 
                    "extractHotelsCodes() failed", ex, baseWebSiteURI);
        }
    }
    
    protected void extractHotelsCodes(Document hotels, ArrayList hotelCodes)
    	throws JaxenException
    {
        assert hotels != null;
        assert hotelCodes != null;
        
		DOMXPath xpath = new DOMXPath(
		        "//DIV[@class='searchResultsList hotelDirectory']/" +
		        "TABLE[1]/TR[@class='odd']");
		
		String code;
		List nodes = xpath.selectNodes(hotels);
		Iterator nodesIter = nodes.iterator();
		Node n;
		int idx;
		while (nodesIter.hasNext())
		{
		    n = (Node)nodesIter.next();
		    n = n.getChildNodes().item(1).getChildNodes().item(1);
		    
		    code = n.getAttributes().getNamedItem("href").getNodeValue();
		    idx = code.lastIndexOf("/");
		    hotelCodes.add(code.substring(idx + 1));
		}
    }
    
    /**
     * @param hotelsLoc The hotels locations to set.
     */
    protected void setHotelsLoc(HotelLocation[] hotelsLocations)
    {
        assert hotelsLocations != null;
        
        hotelsLoc = hotelsLocations;
    }
    
    /**
	 * Checks whether there are more pages with hotels codes to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curLocIdx < hotelsLoc.length;
	}
	
	/**
	 * This method is used only for testing purposes.
	 */
	public void setBaseWebSiteURI(String uri)
	{
	    baseWebSiteURI = uri;
	}
	
	/**
	 * Checks the page (hotels) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document hotels) 
		throws WebSourceParseException, JaxenException
	{
	    if (hotels == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    DOMXPath xpath = new DOMXPath(
		        "//DIV[@class='searchResultsList hotelDirectory']/" +
		        "TABLE[1]/TR[@class='odd']");
	    
	    if (xpath.selectNodes(hotels).size() <= 0)
	    {
	        throw new WebSourceParseException(myAgent.getAID(),
		                "Format of the page is incorrect", 
		                hotels.getDocumentURI());
	    }
	}
}
