package semfly.agents.behaviours;

import jade.core.Agent;

import java.util.ArrayList;

import org.apache.commons.httpclient.methods.PostMethod;
import org.w3c.dom.Document;

import semfly.dto.HotelDocument;
import semfly.exceptions.WebSourceLoadException;
import semfly.net.WebSiteLoader;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for MarriottInfoLoader behaviour.
 * 
 * This class is responsible loading info about Marriott hotels for the hotels
 * codes given from www.marriott.com web site.
 */
public class MarriottInfoLoaderBL
{
    /**
     * The array of hotel's codes that details are to be loaded.
     */
    private String[] hotelCodes;
    
    /**
     * URI address of a web page with Marriott hotels details.
     */
	private String webSiteURI = 
		"http://marriott.com/property/factsheet/";
	
	/**
	 * This const tells us how many pages with hotels details is
	 * loaded befor giving them to the MarriottInfoProcessor behaviour.
	 */
	private final int CodesStep = 4;
	
	/**
	 * Current index of the hotel code in hotelCodes array.
	 */
	private int curCodeIdx = 0;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	/**
	 * Constructor. Sets hotels codes for which all necessary details should 
	 * be loaded (hotelsCodes param) and currend code index on zero.
	 * @param owner Agent - owner of this behaviour.
	 * @param hCodes Marriott hotels codes for which all necessary 
	 * details should be loaded.
	 */
	public MarriottInfoLoaderBL(Agent agent, String[] hCodes)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (hCodes == null) throw new 
	    	IllegalArgumentException("Hotels codes (hCodes) can't be null");
	    
	    myAgent = agent;
		hotelCodes = hCodes;
		curCodeIdx = 0;
	}
	
	/**
	 * Loads details for hotels codes (form hotelCodes array).
	 * @return Loaded web pages with hotels details.
	 * @throws WebSourceLoadException Loading exception.
	 */
	public HotelDocument[] loadResources() throws WebSourceLoadException
	{
		assert hotelCodes != null;
		
		ArrayList docs = new ArrayList();
		int maxIdx = Math.min(curCodeIdx + CodesStep, hotelCodes.length);
		for ( ; curCodeIdx < maxIdx; curCodeIdx++)
		{
			docs.add(new HotelDocument(loadResource(), hotelCodes[curCodeIdx]));
		}
		
		return (HotelDocument[])docs.toArray(new HotelDocument[docs.size()]);
	}
	
	/**
	 * Loads a web page with hotel details.
	 * @return Loaded web resource.
	 */
	protected Document loadResource() throws WebSourceLoadException
	{	
		assert hotelCodes != null;
		assert curCodeIdx >= 0 && curCodeIdx < hotelCodes.length;
		
		return loadHotelDetails(webSiteURI, hotelCodes[curCodeIdx]);
	}
	
	/**
	 * Loads info about Marriott hotels.
	 * @return Loaded web page with info about airports.
	 */
	protected Document loadHotelDetails(String pageURI, String hotelCode)
		throws WebSourceLoadException
	{	
		WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		PostMethod httpMethod = new PostMethod(pageURI + hotelCode);
		
		try
		{
			doc = webLoader.loadWebSiteXHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadHotelDetails failed", ex, pageURI);
		}
		
		return doc;
	}
	
	public String[] getHotelsCodes()
	{
	    return hotelCodes;
	}
	
	/**
	 * Checks whether there are more pages with hotels details to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curCodeIdx < hotelCodes.length;
	}
	
	/**
	 * Sets web site URI. This method is used only for testing purposes.
	 * @param uri
	 */
	public void setWebSiteURI(String uri)
	{
	    webSiteURI = uri;
	}
}
