package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelDocument;
import semfly.dto.HotelInfo;
import semfly.exceptions.WebSourceParseException;
import semfly.onto.jena.Hotel;
import semfly.onto.jena.Location;
import semfly.onto.jena.Vcard;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;

import semfly.textmining.MarriottTextMiner;
import semfly.textmining.StringHelper;
import semfly.textmining.TextDescription;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for MarriottInfoProcessor behaviour.
 *
 * This behaviour parses the DOM objects given with Marriott hotels info 
 * by using Jaxen and then creates an appropriate token to be send to 
 * the Coordinator (by using WrapperSender bahaviour).
 */
public class MarriottInfoProcessorBL
{
    /**
     * Web pages with Marriott hotels details.
     */
	private HotelDocument[] hotels = null;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	private Model hotelModel;
	
	/**
	 * Constructor.
	 * @param owner Agent - owner of this behaviour.
	 * @param marriottDetails Web pages with airports details.
	 */
	public MarriottInfoProcessorBL(Agent agent, HotelDocument[] marriottDetails)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (marriottDetails == null) throw new IllegalArgumentException(
	            "Marriott hotels details (marriottDetails) can't be null");
	   
		hotels = marriottDetails;
		myAgent = agent;
	}
	
	/**
	 * Processes pages with hotels details. Extracts an important info about
	 * hotels and creates an appropriate Jena Model.
	 */
	public Model processResource() throws WebSourceParseException
	{
	    hotelModel = ModelFactory.createDefaultModel();
		
		HotelInfo hotelInfo;
		Resource hotel;
		for (int i = 0; i < hotels.length; i++)
		{
		    hotelInfo = extractHotelBaseInfo(hotels[i].getFactsheetPage());
		    hotelInfo.setHotelCode(hotels[i].getHotelCode().toString());
		    
		    hotel = addHotelInfo2Model(hotelInfo);
		    extractHotelInfo(hotels[i].getFactsheetPage(), hotel);
		}
		
		return hotelModel;
	}
	
	/**
	 * Adds info about a hotel to a Jena Model.
	 * @param hotelInfo Info about a hotel.
	 * @param hotelModel Existing Jena Model.
	 */
	private Resource addHotelInfo2Model(HotelInfo hotelInfo)
	{
	    String hotelURI = Hotel.BaseResourceURI + "Marriott/" + 
	    					hotelInfo.getHotelCode();
	    
		Resource hotel = hotelModel.createResource(hotelURI);
		hotel.addProperty(Hotel.id, "Marriott/" + hotelInfo.getHotelCode());
		//hotel.addProperty(Hotel.) //TODO: title
		
		//address
		Resource address = hotelModel.createResource(hotelURI + "/Address");
		address.addProperty(Vcard.Street, hotelInfo.getStreet());
		address.addProperty(Vcard.Locality, hotelInfo.getLocality());
		address.addProperty(Vcard.Pcode, hotelInfo.getPcode());
		address.addProperty(Vcard.Country, hotelInfo.getCountry());
		hotel.addProperty(Location.address, address);
		
		//phone and fax
		hotel.addProperty(Location.phone, hotelInfo.getPhone());
		
		return hotel;
	}
	
	/**
	 * Extracts info about a hotel from the web page given.
	 * @param hotelDoc Web page with hotel details.
	 * @return Object with an important info about a hotel.
	 */
	private HotelInfo extractHotelBaseInfo(Document hotelDoc) 
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	   
		HotelInfo hotelInfo = new HotelInfo();
		
		extractHotelAddress(hotelDoc, hotelInfo);
	
		return hotelInfo;
	}
	
	private void extractHotelAddress(Document hotelDoc, HotelInfo hotelInfo)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//DIV[@id='hotelAddressSub']/H2[1]/text()");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			hotelInfo.setName(node.getNodeValue());
			
			xpath = 
		        new DOMXPath("//DIV[@id='hotelAddressSub']/UL[1]/LI");
			List nodes = xpath.selectNodes(hotelDoc);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getFirstChild().getNodeValue();
			   
			    parseHotelAddress(nodeValue, idx++, hotelInfo);
			}
		}
		catch (Exception ex)
		{
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseHotelAddress(String nodeValue, int idx, HotelInfo hotelInfo)
	{
	    if (idx == 0) //street
	    {
		    nodeValue = nodeValue.replaceAll("\n", "");
		    nodeValue = nodeValue.replaceAll("\t", "");
		    nodeValue = nodeValue.trim();
		    
		    hotelInfo.setStreet(nodeValue);
	    }
	    else if (idx == 1) //locality, country, pcode
	    {
	        int l = nodeValue.lastIndexOf(",");
	        String locality = nodeValue.substring(0, l);
	        locality = locality.replaceAll("\n", "");
	        locality = locality.replaceAll("\t", "");
	        hotelInfo.setLocality(locality);
	        
	        String[] parts = nodeValue.split("\n\t");
	        ArrayList txt = new ArrayList();
	        for (int j = 0; j < parts.length; j++)
	        {
	            parts[j] = parts[j].trim();
	            if (parts[j].length() > 1) txt.add(parts[j]); 
	        }
	        parts = (String[])txt.toArray(new String[txt.size()]);
	        
	        String pcode = "";
	        for (int j = 1; j < parts.length - 1; j++)
	        {
	            pcode += parts[j];
	            if (j < parts.length - 2) pcode += " ";
	        }
	        hotelInfo.setPcode(pcode);
	        
	        hotelInfo.setCountry(parts[parts.length - 1]);
	    }
	    else if (idx == 2) //phone
	    {
	        nodeValue = nodeValue.replaceAll("\n", "");
		    nodeValue = nodeValue.replaceAll("\t", "");
		    nodeValue = nodeValue.trim();
		    
	        int i = nodeValue.indexOf(":");
	        nodeValue = nodeValue.substring(i + 1).trim();
	        hotelInfo.setPhone(nodeValue.substring(1));
	    }
	    else if (idx == 3) //fax
	    {
	        nodeValue = nodeValue.replaceAll("\n", "");
		    nodeValue = nodeValue.replaceAll("\t", "");
		    nodeValue = nodeValue.trim();
		    
	        int i = nodeValue.indexOf(":");
	        nodeValue = nodeValue.substring(i + 1).trim();
	        hotelInfo.setFax(nodeValue.substring(1));
	    }
	}
	
	private void extractHotelInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    extractRoomAmenityInfo(hotelDoc, hotelRes);
	    extractRecreationInfo(hotelDoc, hotelRes);
	    extractPetPolicyInfo(hotelDoc, hotelRes);
	    extractHotelDetailsInfo(hotelDoc, hotelRes);
	    extractCheckInCheckOutInfo(hotelDoc, hotelRes);
	}
	
	private void extractRecreationInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//A[@name='recreation']");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			
			xpath = new DOMXPath("following-sibling::DIV[position()=1]/" +
					"descendant::LI/text()");
			
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			
			ArrayList recreation = new ArrayList();
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    if (nodeValue.contains("Golf")) nodeValue = "Golf";
			    else
			    {
			        idx = nodeValue.indexOf("(");
			        if (idx > 0) nodeValue = nodeValue.substring(0, idx);
			    }
			    if (nodeValue.length() > 0) recreation.add(nodeValue);
			}
			
			parseRecreationInfo(
		            (String[])recreation.toArray(new String[recreation.size()]), 
		            	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseRecreationInfo(String[] recreation, Resource hotelRes)
	{
	    MarriottTextMiner txtMiner = MarriottTextMiner.getInstance();
	    TextDescription txtDesc;
	    for (int i = 0; i < recreation.length; i++)
	    {
	        txtDesc = txtMiner.getRecreationDescription(recreation[i]);
	        if (txtDesc != null)
	        {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
	        }
	    }
	}
	
	private void extractRoomAmenityInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//A[@name='guestroom']");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			
			xpath = new DOMXPath("following-sibling::DIV[position()=1]/" +
					"descendant::LI/text()");
			
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList roomInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    roomInfo.add(nodeValue);
			}
			
			parseRoomAmenityInfo(
		            (String[])roomInfo.toArray(new String[roomInfo.size()]), 
		            	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseRoomAmenityInfo(String[] amenities, Resource hotelRes)
	{
	    MarriottTextMiner txtMiner = MarriottTextMiner.getInstance();
	    TextDescription txtDesc;
	    for (int i = 0; i < amenities.length; i++)
	    {
	        txtDesc = txtMiner.getAmenityDescription(amenities[i]);
	        if (txtDesc != null)
	        {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
	        }
	    }
	}
	
	private void extractPetPolicyInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//H3");
		
			List nodes = xpath.selectNodes(hotelDoc);
			Iterator nodesIter = nodes.iterator();    
			Node node; 
			String nodeValue;
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getFirstChild().getNodeValue();
			    nodeValue = nodeValue.replace(" ", "");
			    if (nodeValue.equalsIgnoreCase("PetPolicy"))
			    {
				    xpath = new DOMXPath("following-sibling::UL/LI/text()");
				    node = (Node)xpath.selectSingleNode(node);
				    
				    nodeValue = node.getNodeValue();
				    nodeValue = nodeValue.replace("\t", "");
				    nodeValue = nodeValue.replace("\n", "");
				    nodeValue = nodeValue.replace(" ", "");
				    
				
				    MarriottTextMiner txtMiner = MarriottTextMiner.getInstance();
				    TextDescription txtDesc;
				    txtDesc = txtMiner.getPetsPolicyDescription(nodeValue);
				    
				    if (txtDesc != null)
				    {
				        hotelRes.addProperty(txtDesc.getProperty(), 
			                    txtDesc.getResource());
				    }
			        
			        break;
			    }
			}
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void extractHotelDetailsInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
	        DOMXPath xpath = 
		        new DOMXPath("//A[@name='details']");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			
			xpath = new DOMXPath("following-sibling::UL[position()=1]/" +
					"descendant::LI/text()");
			
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			   
			    parseHotelDetailsInfo(nodeValue, idx++, hotelRes);
			}
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelDetailsInfo() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseHotelDetailsInfo(String nodeValue, int idx, 
	        							Resource hotelRes)
	{
	    StringHelper stringHelper = StringHelper.getInstance();
	    MarriottTextMiner txtMiner = MarriottTextMiner.getInstance();
	    TextDescription txtDesc;
	    
	    nodeValue = nodeValue.replace(" ", "");
        nodeValue = nodeValue.replace("\t", "");
        nodeValue = nodeValue.replace("\n", "");
	    if (idx == 0) //x floors, y rooms, z suites
	    {
	        String floors = null, rooms = null, suites = null;
	        String[] info = nodeValue.split(",");
	        
	        if (info.length > 0) floors = info[0];
	        if (info.length > 1) rooms = info[1];
	        if (info.length > 2) suites = info[2];
	        
	        Long quantity;
	        long totalRoomsAndSuites = 0;
	        if (floors != null)
	        {
	            quantity = stringHelper.extractPrefixNumber(floors);
		        if (quantity != null) //floors
		        {
		            Resource floorsRes = hotelModel.createResource(
		                    hotelRes.getURI() + "/Floors");
		            floorsRes.addProperty(Hotel.quantity, quantity.longValue());
		            hotelRes.addProperty(Hotel.roomInfo, floorsRes);
		        }
	        }
	        if (rooms != null)
	        {
		        quantity = stringHelper.extractPrefixNumber(rooms);
		        if (quantity != null) //rooms
		        {
		            Resource roomsRes = hotelModel.createResource(
		                    hotelRes.getURI() + "/TotalRooms");
		            roomsRes.addProperty(Hotel.quantity, quantity.longValue());
		            hotelRes.addProperty(Hotel.roomInfo, roomsRes);
		            
		            totalRoomsAndSuites += quantity.longValue();
		        }
	        }
	        if (suites != null)
	        {
		        quantity = stringHelper.extractPrefixNumber(suites);
		        if (quantity != null) //suites
		        {
		            Resource suitesRes = hotelModel.createResource(
		                    hotelRes.getURI() + "/Suites");
		            suitesRes.addProperty(Hotel.quantity, quantity.longValue());
		            hotelRes.addProperty(Hotel.roomInfo, suitesRes);
		            
		            totalRoomsAndSuites += quantity.longValue();
		        }
	        }
	        
	        //total rooms and suites
	        if (totalRoomsAndSuites > 0)
	        {
	            Resource totalRSRes = hotelModel.createResource(
	                    hotelRes.getURI() + "/TotalRoomsAndSuites");
	            totalRSRes.addProperty(Hotel.quantity, totalRoomsAndSuites);
	            hotelRes.addProperty(Hotel.roomInfo, totalRSRes);
		    }   
	    }
	    else //x The Room That Works guest rooms
	    {
	        if (nodeValue.indexOf("TheRoomThatWorksguestrooms") >= 0)
	        {
	            Long quantity = stringHelper.extractPrefixNumber(nodeValue);
	            if (quantity != null) //suites
		        {
	                Resource worksRoomRes = hotelModel.createResource(
		                    hotelRes.getURI() + "/TheRoomThatWorksGuestRooms");
		            worksRoomRes.addProperty(Hotel.quantity, quantity.longValue());
		            hotelRes.addProperty(Hotel.roomInfo, worksRoomRes);
		        }
	        }
	    }
	    
	}
	
	private void extractCheckInCheckOutInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//H3");
		
			List nodes = xpath.selectNodes(hotelDoc);
			Iterator nodesIter = nodes.iterator();    
			Node node; 
			String nodeValue;
			StringBuffer checkBuffer = new StringBuffer();
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getFirstChild().getNodeValue();
			    nodeValue = nodeValue.replace(" ", "");
			    if (nodeValue.equalsIgnoreCase("Check-InandCheckout"))
			    {
				    xpath = new DOMXPath("following-sibling::UL/LI/text()");
				    nodes = xpath.selectNodes(node);
				    nodesIter = nodes.iterator();    
				    while (nodesIter.hasNext())
					{
				        node = (Node)nodesIter.next();
				       
				        nodeValue = node.getNodeValue();
					    nodeValue = nodeValue.replace("\t", "");
					    nodeValue = nodeValue.replace("\n", " ");
					    nodeValue = nodeValue.replace("  ", " ");
					    nodeValue = nodeValue.replace("  ", " ");
					    nodeValue = nodeValue.trim();
					    
					    if ( !nodeValue.equals("") )
					    {
						    checkBuffer.append(nodeValue);
						    checkBuffer.append("; ");
					    }
					}
				    createCheckInCheckOutDescription(checkBuffer.toString(),
				            	hotelRes);
				    
			        break;
			    }
			}
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void createCheckInCheckOutDescription(String info, Resource hotelRes)
	{
	    String checkInOut;
	    info = info.trim();
	    int idx = info.indexOf(";");
	    
	    if (idx < 0) checkInOut = info;
	    else
	    {
	        checkInOut = info.substring(0, idx + 1);
	        checkInOut += " ";
	        checkInOut += info.substring(idx + 2, info.length());
	    }
	    
	    Resource checkRes = hotelModel.createResource(
	            hotelRes.getURI() + "/CheckIn-CheckOut");
	    checkRes.addProperty(Hotel.additionalDetail, checkInOut);
	    
	    hotelRes.addProperty(Hotel.additionalDetail, checkRes);
	}
}
