package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelLocation;
import semfly.dto.WebPageDocument;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;

/**
 * @author szymek
 *
 * This class plays as a business logic for MarriottLocationsLoader behaviour.
 * 
 * This class represents a Behaviour that is to load locations (their codes) 
 * of all Marriot Hotels. 
 * Starting page is: "http://marriott.com/search/hotelDirectory.mi?
 * country=PL&bc=&hotelDirectory=true";
 */
public class MarriottLocationsLoaderBL
{
    private String pageURI = "http://marriott.com/search/" +
    "hotelDirectory.mi?country=PL&bc=&hotelDirectory=true";

	private final String countryQS = "country=",
							usStateQS = "state=";
	
	private Agent myAgent;
	
	public MarriottLocationsLoaderBL(Agent owner)
	{
	    myAgent = owner;
	}
	
	/**
	 * Loads locations of Marriott Hotels.
	 * @return
	 * @throws WebSourceLoadException
	 */
	public WebPageDocument loadLocations() throws WebSourceLoadException
	{
	    WebSiteLoader webLoader = new WebSiteLoader();
		Document doc = null;
		
		GetMethod httpMethod = new GetMethod(pageURI);
		try
		{
			doc = webLoader.loadWebSiteXHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadLocations() failed", ex, pageURI);
		}
		
		return new WebPageDocument(doc);
	}
	
	/**
	 * Extracts locations codes from a web page with hotels locations.
	 * @param locations
	 * @return
	 * @throws WebSourceParseException
	 */
	public HotelLocation[] extractLocationCodes(Document locations) 
		throws WebSourceParseException
	{
	    if (locations == null)
	        throw new IllegalArgumentException(
	                "Page with hotels locations can't be null");
	    
	    try
	    {
	        checkPageFormat(locations);
	        
	        ArrayList hotels = new ArrayList();
	        extractCountryLocations(locations, hotels);
	        extractUSStateLocations(locations, hotels);
	        
	        return (HotelLocation[])hotels.toArray(
	                	new HotelLocation[hotels.size()]);
	    }
	    catch (WebSourceParseException ex)
		{
		    throw ex;
		}
	    catch (Exception ex)
	    {
	        throw new WebSourceParseException(myAgent.getAID(), 
			        "extractLocationCodes() failed", ex, pageURI);
	    }
	}
	
	/**
	 * Extracts locations codes for countries from a web page 
	 * with hotels locations.
	 * @param locations
	 * @param hotels
	 * @throws JaxenException
	 */
	private void extractCountryLocations(Document locations, ArrayList hotels)
		throws JaxenException
	{
	    assert locations != null;
	    assert hotels != null;
	    
		DOMXPath xpath = new DOMXPath("//SELECT[@name='countrySelect']/" +
				"OPTION[@value != ''");
		
		String code;
		List nodes = xpath.selectNodes(locations);
		Iterator nodesIter = nodes.iterator();
		while (nodesIter.hasNext())
		{
		    code = extractLocationCode((Node)nodesIter.next());
		    hotels.add(new HotelLocation(code, HotelLocation.Country));
		}
	}
	
	/**
	 * Extracts locations codes for states in the USA from a web page 
	 * with hotels locations.
	 * @param locations
	 * @param hotels
	 * @throws JaxenException
	 */
	private void extractUSStateLocations(Document locations, ArrayList hotels)
		throws JaxenException
	{
	    assert locations != null;
	    assert hotels != null;
	   
		DOMXPath xpath = new DOMXPath("//SELECT[@name='stateSelect']/" +
				"OPTION[@value != ''");
		
		String code;
		List nodes = xpath.selectNodes(locations);
		Iterator nodesIter = nodes.iterator();
		while (nodesIter.hasNext())
		{
		    code = extractLocationCode((Node)nodesIter.next());
		    hotels.add(new HotelLocation(code, HotelLocation.USState));
		}
	}
	
	/**
	 * Extracts location code from the node given.
	 * @param node
	 * @return
	 */
	private String extractLocationCode(Node node)
	{
	    assert node != null;
	    
	    String queryStr = node.getAttributes().
	    						getNamedItem("value").getNodeValue();
	    
	    int startIdx = queryStr.indexOf(countryQS);
	    int codeNameLen = countryQS.length();
	    
	    if (startIdx < 0) 
	    {
	        startIdx = queryStr.indexOf(usStateQS);
	        codeNameLen = usStateQS.length();
	    }
	        
	    String code = queryStr.substring(startIdx);
	    int endIdx = code.indexOf("&");
	   
	    return code.substring(codeNameLen, endIdx);
	}
	
	/**
	 * Checks the page (locations) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document locations) 
		throws WebSourceParseException, JaxenException
	{
	    if (locations == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    DOMXPath xpath = new DOMXPath("//SELECT[@name='countrySelect']/" +
		"OPTION[@value != ''");  
	    
	    if (xpath.selectNodes(locations).size() <= 0)
	    {
	        xpath = new DOMXPath("//SELECT[@name='stateSelect']/" +
							"OPTION[@value != ''");
	        
	        if (xpath.selectNodes(locations).size() <= 0)
		    {
	            throw new WebSourceParseException(myAgent.getAID(),
		                "Format of the page is incorrect", 
		                locations.getDocumentURI());
		    }
	    }
	}
	
	/**
	 * This method is used only for testing purposes.
	 * @param uri
	 */
	public void setPageURI(String uri)
	{
	    pageURI = uri;
	}
}
