package semfly.agents.behaviours;

import jade.core.Agent;
import jade.core.behaviours.OneShotBehaviour;
import jade.lang.acl.ACLMessage;
import jade.core.AID;
import java.io.StringWriter;
import java.util.Date;
import jade.domain.FIPANames;
import semfly.log.LogWriter;
import semfly.onto.jade.IndexToken;
import semfly.onto.jade.SaveInTPQ;
import semfly.agents.CoordinatorAgent;
import semfly.agents.WrapperAgent;
import semfly.exceptions.ExceptionManager;
import semfly.globals.DFAdapter;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.content.lang.Codec;
import com.hp.hpl.jena.rdf.model.Model;
import semfly.exceptions.WebSourceParseException;


/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This abstract class is a base for all classes that are responsible for
 * processing resources. Owner of this class must be a Wrapper Agent.
 */
public abstract class ResourceProcessor extends OneShotBehaviour
{
    /**
     * Timeout for Contract-Net-Protocol.
     */
	protected final int ContractTimeout = 1000;
	
	/**
	 * Jena Model of processed resource.
	 */
	protected Model model;
	
	protected ResourceProcessor()
	{
		super();
	}
	
	protected ResourceProcessor(Agent owner)
	{
		super(owner);
		assert owner instanceof WrapperAgent;
	}

	/**
	 * Processes a resource and then sends result to the Coordinator.
	 * @see jade.core.behaviours.Behaviour#action()
	 */
	public void action()
	{
		assert model != null;
		
		try
		{
			processResource();
			if (!model.isEmpty()) 
		    {
			    sendResourceMetadata();
			    ((WrapperAgent)myAgent).incNumberOfVisitedPages();
		    }
		}
		catch (Exception ex)
		{
		    ExceptionManager.publish(ex);
		}
	}
	
	public int onEnd()
	{
		return 0;
	}
	
	/**
	 * Abstract method. Processes resource.
	 *
	 */
	protected abstract void processResource() throws WebSourceParseException;
	
	/**
	 * Sends metadata about just processed resource to the Coordinator.
	 * This method adds to the owner's behaviours 
	 * ResourceMetadataSender behaviour which participate in 
	 * Contract Net Protocol (is the Contract Net Protocol Initiator).
	 * Calls prepareALCMessage for creating ACLMessage object which is the
	 * parameter for ResourceMetadataSender constructor.
	 */
	protected void sendResourceMetadata()
	{
		assert model != null;
		
		ACLMessage msg = null;
		try
		{
		     msg = prepareACLMessage();
		}
		catch (Exception ex)
		{
		    ExceptionManager.publish(WrapperAgent.ServiceType, ex, 
		            "ACL message preparation failed");
		}
		
		if (msg != null)
		{
			((WrapperAgent)myAgent).addBusinessSubBehaviour(
						new WrapperSender(myAgent, msg));
		}
	}
	
	/**
	 * Prepares ACL message for being a parameter for the ContractNetInitiator 
	 * contsructor. 
	 * If there isn't any receiver (CoordinatorAgent) returns null and
	 * writes info about it to the CCS log.
	 * @return Created ACL message or null if there isn't any receiver.
	 */
	protected ACLMessage prepareACLMessage() throws OntologyException, 
		Codec.CodecException
	{
		assert model != null;
		
  		AID[] coordinators = searchDF(CoordinatorAgent.ServiceType);
  		
  		if (coordinators != null && coordinators.length > 0)
  		{
  		    return fillMessageContent(coordinators);
	  	}
  		else
  		{
  		    LogWriter.writeWarn2Log(WrapperAgent.ServiceType, 
  		            "There is no receiver", model);
  			return null;
  		}
	}
	
	/**
	 * Fills ACL message content (message includes an IndexToken).
	 * 
	 * @param coordinators Message receivers.
	 * @return Created message.
	 * @throws OntologyException
	 * @throws Codec.CodecException
	 */
	private ACLMessage fillMessageContent(AID[] coordinators) 
		throws OntologyException, Codec.CodecException
	{
	    assert coordinators != null;
	    
	    ACLMessage msg = new ACLMessage(ACLMessage.CFP);
  		msg.setLanguage(((WrapperAgent)myAgent).getLanguageName());
  		msg.setOntology(((WrapperAgent)myAgent).getOntologyName());
  		
	    for (int i = 0; i < coordinators.length; i++)
  			msg.addReceiver(coordinators[i]);
  		
  		msg.setProtocol(FIPANames.InteractionProtocol.FIPA_CONTRACT_NET);
  		msg.setReplyByDate(
  				new Date(System.currentTimeMillis() + ContractTimeout));
  		
  		Action contentAction = createAgentAction();
  		
  		myAgent.getContentManager().fillContent(msg, contentAction);
  		return msg;
	}
	
	/**
	 * Creates AgentAction - SaveInTPQ.
	 * @return
	 */
	private Action createAgentAction()
	{
		IndexToken token = new IndexToken();
  		token.setCreatedBy(myAgent.getAID());
  		token.setCreatedDate(new Date());
  		token.setJenaModel(getModelAsString(model));
  		
  		SaveInTPQ saveAction = new SaveInTPQ();
  		saveAction.setIndexToken(token);
  		
  		Action contentAction = new Action();
  		contentAction.setAction(saveAction);
  		contentAction.setActor(myAgent.getAID());
  		
		return contentAction;
	}
	
	/**
	 * Returns the model given as a string.
	 * @param m
	 * @return
	 */
	private String getModelAsString(Model m)
	{
		assert m != null;
		
		StringWriter writer = new StringWriter();
		m.write(writer);
		
		return writer.toString();
	}

	/**
	 * Search all agents based on the service type given by means of DFAdapter
	 * class.
	 * @param service Service type.
	 * @return AIDs of found agents of the service given, null if no agent 
	 * has been found or an exception has occured.
	 */
	protected AID[] searchDF(String service)
	{
		DFAdapter searcher = DFAdapter.getInstance();
		
		return searcher.searchDF(myAgent, service);
	}
}
