package semfly.agents.behaviours;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import semfly.dto.ResourceDocument;
import semfly.dto.SheratonHotelCode;
import semfly.dto.WebPageDocument;
import org.apache.commons.httpclient.methods.GetMethod;
import org.jaxen.JaxenException;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import jade.core.Agent;
import semfly.dto.HotelLocation;
import semfly.exceptions.WebSourceLoadException;
import semfly.exceptions.WebSourceParseException;
import semfly.net.WebSiteLoader;


/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class is responsible for loading Sheraton Hotels Codes.
 */
public class SheratonHotelCodesLoaderBL
{
    private String firstBaseWebSiteURI = 
		"http://www.starwoodhotels.com/sheraton/go.search?" + 
			"searchType=city&country=";
    
    private String secondBaseWebSiteURI = 
        "http://www.starwoodhotels.com/sheraton/search/search_results.html?" + 
        "sortNum=999&requestType=sort&page=&sortType=relevance&country=";
    
    private final String USStateProvinceQS = "stateProvince";
    
    private final String USACode = "US";
	
	/**
	 * This const tells us how many pages with hotel codes is
	 * loaded befor giving them to the SheratonInfoLoader behaviour.
	 */
	private final int CodesStep = 4;
	
	/**
	 * Array with hotels locations for which info must be loaded.
	 */
	private HotelLocation[] hotelsLoc;
	
	/**
	 * Current index of the hotel location in hotelsLoc array.
	 */
	private int curLocIdx = 0;
	
	private Agent myAgent;
   
    
    public SheratonHotelCodesLoaderBL(Agent agent, 
            	HotelLocation[] hotelsLocations)
    {
        myAgent = agent;
        hotelsLoc = hotelsLocations;
    }
    
    /**
     * Loads pages with Sheraton hotels codes.
     * @return Loaded web pages.
     * @throws WebSourceLoadException
     */
	public WebPageDocument[] loadResources() throws WebSourceLoadException
	{
	    ArrayList docs = new ArrayList();
		
	    int maxIdx = Math.min(curLocIdx + CodesStep, hotelsLoc.length);
		for ( ; curLocIdx < maxIdx; curLocIdx++)
		{
			docs.add(loadResource());
		}
		
		return (WebPageDocument[])docs.toArray(new WebPageDocument[docs.size()]);
	}
	
    /** (non-Javadoc)
     * @see semfly.agents.behaviours.ResourceLoader#loadResource()
     */
    protected ResourceDocument loadResource() throws WebSourceLoadException
    {
        assert hotelsLoc != null;
        assert curLocIdx >= 0 && curLocIdx < hotelsLoc.length;
        
        HotelLocation hotelLoc = hotelsLoc[curLocIdx];
        String firstURI = firstBaseWebSiteURI,
        		secondURI = secondBaseWebSiteURI; 
    
        if (hotelLoc.isLocatedInUSState())
        {
            firstURI += USACode + "&" + USStateProvinceQS + 
            			"=" + hotelLoc.getLocationCode();
            secondURI += USACode + "&" + USStateProvinceQS + 
						"=" + hotelLoc.getLocationCode();
        }
        else 
        {
            firstURI += hotelLoc.getLocationCode();
            secondURI += hotelLoc.getLocationCode();
        }
        
        return new WebPageDocument(loadHotelsPage(firstURI, secondURI));
    }
    
    protected Document loadHotelsPage(String firstURI, String secondURI) 
		throws WebSourceLoadException
	{
	    WebSiteLoader webLoader = new WebSiteLoader();
		GetMethod httpMethod = new GetMethod(firstURI);
		
		try
		{
		    webLoader.loadWebSiteHTML(httpMethod);
		    httpMethod = new GetMethod(secondURI);
		    
		    return webLoader.loadWebSiteHTML(httpMethod);
		}
		catch (Exception ex)
		{
			throw new WebSourceLoadException(myAgent.getAID(), 
			        "loadHotelsPage() failed", ex, firstURI);
		}
	}
    
    /**
     * Extracts hotels codes from the array with web pages given.
     * @param hotels An array of web pages with hotels codes.
     * @return Extracted codes.
     * @throws JaxenException
     */
    public SheratonHotelCode[] extractHotelsCodes(WebPageDocument[] hotels) 
    	throws WebSourceParseException
    { 
        String page = "";
        try
        {
	        ArrayList hotelsCodes = new ArrayList();
	        for (int i = 0; i < hotels.length; i++)
	        {
	            checkPageFormat(hotels[i].getWebDocument());
	            extractHotelsCodes(hotels[i].getWebDocument(), hotelsCodes);
	        }
	        
	        return (SheratonHotelCode[])hotelsCodes.toArray(
	                	new SheratonHotelCode[hotelsCodes.size()]);
	    }
        catch (WebSourceParseException ex) 
        {
            throw ex;
        }
        catch (JaxenException ex)
        {
            throw new WebSourceParseException(myAgent.getAID(), 
                    "extractHotelsCodes() failed", ex, firstBaseWebSiteURI);
        }
    }
    
    protected void extractHotelsCodes(Document hotels, ArrayList hotelCodes)
    	throws JaxenException
    {
        assert hotels != null;
        assert hotelCodes != null;
        
		DOMXPath xpath = new DOMXPath("//DIV[@class='propertyContainer']");
		
		String hotelLink;
		List nodes = xpath.selectNodes(hotels);
		Iterator nodesIter = nodes.iterator();
		Node n;
		int idx;
		while (nodesIter.hasNext())
		{
		    n = (Node)nodesIter.next();
		    
		    n = n.getChildNodes().item(1);
		    n = n.getChildNodes().item(1);
		    n = n.getChildNodes().item(1);
		    n = n.getChildNodes().item(1);
		    
		    hotelLink = n.getAttributes().getNamedItem("href").getNodeValue();
		    hotelCodes.add(createHotelCode(hotelLink));
		}
    }
    
    private SheratonHotelCode createHotelCode(String hotelLink)
    {
        SheratonHotelCode hotelCode = new SheratonHotelCode();
        
        int idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setPropertyID(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
        
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setBrand(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
        
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setChainCode(hotelLink.substring(0, idx));
        hotelLink = hotelLink.substring(idx + 1);
       
        idx = hotelLink.indexOf("'");
        hotelLink = hotelLink.substring(idx + 1);
        idx = hotelLink.indexOf("'");
        hotelCode.setAffilationCode(hotelLink.substring(0, idx));
        
        return hotelCode;
    }
    
    /**
	 * Checks the page (hotels) given format. If the format is incorrect,
	 * WebSourceParseException is thrown.
	 * 
	 * @throws WebSourceParseException
	 */
	private void checkPageFormat(Document hotels) 
		throws WebSourceParseException, JaxenException
	{
	    if (hotels == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	    
	    DOMXPath xpath = new DOMXPath("//DIV[@class='propertyContainer']");
	    
	    if (xpath.selectNodes(hotels).size() <= 0)
	    {
	        xpath = new DOMXPath("//TD[@class='error']");

			if (xpath.selectNodes(hotels).size() <= 0)
			{
			    throw new WebSourceParseException(myAgent.getAID(),
	                	hotels.getDocumentURI(),
		                "Format of the page is incorrect");
			}
		}
	}
	
	/**
	 * Checks whether there are more pages with hotels codes to load.
	 * @return True - there are more pages, otherwise False.
	 */
	public boolean morePages2Load()
	{
	    return curLocIdx < hotelsLoc.length;
	}
	
	/**
	 * Sets first base URI - this method is used only for testing purposes.
	 */
    public void setFirstBaseWebSiteURI(String firstBaseWebSiteURI)
    {
        this.firstBaseWebSiteURI = firstBaseWebSiteURI;
    }
    
    /**
	 * Sets second base URI - this method is used only for testing purposes.
	 */
    public void setSecondBaseWebSiteURI(String secondBaseWebSiteURI)
    {
        this.secondBaseWebSiteURI = secondBaseWebSiteURI;
    }
}
