package semfly.agents.behaviours;

import jade.core.Agent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import semfly.dto.HotelDocument;
import semfly.dto.HotelInfo;
import semfly.exceptions.WebSourceParseException;
import semfly.onto.jena.Hotel;
import semfly.onto.jena.Location;
import semfly.onto.jena.Vcard;
import semfly.textmining.SheratonTextMiner;
import semfly.textmining.TextDescription;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for SheratonInfoProcessor behaviour.
 *
 * This behaviour parses the DOM objects given with Sheraton hotels info 
 * by using Jaxen and then creates an appropriate token to be send to 
 * the Coordinator (by using WrapperSender bahaviour).
 */
public class SheratonInfoProcessorBL
{
    /**
     * Web pages with Sheraton hotels details.
     */
	private HotelDocument[] hotels = null;
	
	/**
	 * Agent who's using this class.
	 */
	private Agent myAgent;
	
	private Model hotelModel;
	
	/**
	 * Constructor.
	 * @param owner Agent - owner of this behaviour.
	 * @param sheratonDetails Web pages with airports details.
	 */
	public SheratonInfoProcessorBL(Agent agent, HotelDocument[] sheratonDetails)
	{
	    if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
	    if (sheratonDetails == null) throw new IllegalArgumentException(
	            "Sheraton hotels details (sheratonDetails) can't be null");
	   
		hotels = sheratonDetails;
		myAgent = agent;
	}
	
	/**
	 * Processes pages with hotels details. Extracts an important info about
	 * hotels and creates an appropriate Jena Model.
	 */
	public Model processResource() throws WebSourceParseException
	{
	    hotelModel = ModelFactory.createDefaultModel();
		
		HotelInfo hotelInfo;
		Resource hotel;
		for (int i = 0; i < hotels.length; i++)
		{
		    hotelInfo = extractHotelBaseInfo(hotels[i].getAmenitiesPage());
		    hotelInfo.setHotelCode(hotels[i].getHotelCode().toString());
		    
		    hotel = addHotelInfo2Model(hotelInfo, hotelModel);
		    extractHotelInfo(hotels[i], hotel);
		}
		
		return hotelModel;
	}
	
	/**
	 * Adds info about a hotel to a Jena Model.
	 * @param hotelInfo Info about a hotel.
	 * @param hotelModel Existing Jena Model.
	 */
	private Resource addHotelInfo2Model(HotelInfo hotelInfo, Model hotelModel)
	{
	    String hotelURI = Hotel.BaseResourceURI + "Sheraton/" + 
	    					hotelInfo.getHotelCode();
	    
		Resource hotel = hotelModel.createResource(hotelURI);
		hotel.addProperty(Hotel.id, "Sheraton/" + hotelInfo.getHotelCode());
		
		//address
		Resource address = hotelModel.createResource(hotelURI + "/Address");
		String propValue;
		
		propValue = hotelInfo.getStreet();
		if (propValue != null && propValue.length() > 0)
		    address.addProperty(Vcard.Street, propValue);
		
		propValue = hotelInfo.getLocality();
		if (propValue != null && propValue.length() > 0)
		    address.addProperty(Vcard.Locality, propValue);
		
		//address.addProperty(Vcard.Pcode, hotelInfo.getPcode());
		
		propValue = hotelInfo.getCountry();
		if (propValue != null && propValue.length() > 0)
		    address.addProperty(Vcard.Country, propValue);
		
		hotel.addProperty(Location.address, address);
		
		//phone and fax
		propValue = hotelInfo.getPhone();
		if (propValue != null && propValue.length() > 0)
		    hotel.addProperty(Location.phone, propValue);
		
		return hotel;
	}
	
	/**
	 * Extracts info about a hotel from the web page given.
	 * @param hotelDoc Web page with hotel details.
	 * @return Object with an important info about a hotel.
	 */
	private HotelInfo extractHotelBaseInfo(Document hotelDoc) 
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	   
	    HotelInfo hotelInfo = new HotelInfo();
	    try
	    {
			
			
			extractHotelAddress(hotelDoc, hotelInfo);
			
			return hotelInfo;
	    }
	    catch (Exception ex)
	    {
	        ex.printStackTrace();
	    }
	    
	    return hotelInfo;
	}
	
	/**
	 * Extracts info about a hotel.
	 * @param hotelDoc Web pages with hotel details.
	 * @return Object with an important info about a hotel.
	 */
	private void extractHotelInfo(HotelDocument hotelDoc, Resource hotelRes) 
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Documents to parse can't be null");
	
	    extractRoomAmenityInfo(hotelDoc.getAmenitiesPage(), hotelRes);
		extractHotelServices(hotelDoc.getServicesPage(), hotelRes);
		extractCheckInCheckOutInfo(hotelDoc.getHomePage(), hotelRes);
	}
	
	private void extractRoomAmenityInfo(Document hotelDoc, 
	        		Resource hotelRes) throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//DIV[@id='guestRoomsTitle']");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			
			xpath = new DOMXPath("ancestor::TR[position()=1]/" +
					"following-sibling::TR[position()=2]/descendant::LI/text()");
			
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList roomInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    roomInfo.add(nodeValue);
			}
			
			parseRoomAmenityInfo(
		            (String[])roomInfo.toArray(new String[roomInfo.size()]), 
		            	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseRoomAmenityInfo(String[] amenities, Resource hotelRes)
	{
	    SheratonTextMiner txtMiner = SheratonTextMiner.getInstance();
	    TextDescription txtDesc;
	    for (int i = 0; i < amenities.length; i++)
	    {
	        txtDesc = txtMiner.getAmenityDescription(amenities[i]);
	        if (txtDesc != null)
	        {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
	        }
	    }
	}
	
	private void extractHotelServices(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    if (hotelDoc == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//DIV[@id='hotelServicesTitle']");
		
			Node node = (Node)xpath.selectSingleNode(hotelDoc);
			if (node == null) return;
			
			xpath = new DOMXPath("ancestor::TR[position()=1]/" +
					"following-sibling::TR[position()=2]/descendant::LI/text()");
			
			List nodes = xpath.selectNodes(node);
			Iterator nodesIter = nodes.iterator();
			int idx = 0;
			String nodeValue;
			
			if (nodes.size() <= 0) return;
			    
			ArrayList servicesInfo = new ArrayList(); 
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();
			    nodeValue = nodeValue.replace("\t", "");
			    nodeValue = nodeValue.replace("\n", "");
			    nodeValue = nodeValue.replace(" ", "");
			    
			    servicesInfo.add(nodeValue);
			}
			
			parseHotelServicesInfo(
		            (String[])servicesInfo.toArray(new String[servicesInfo.size()]), 
		            	hotelRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
	
	private void parseHotelServicesInfo(String[] services, Resource hotelRes)
	{
	    SheratonTextMiner txtMiner = SheratonTextMiner.getInstance();
	    TextDescription txtDesc;
	    for (int i = 0; i < services.length; i++)
	    {
	        txtDesc = txtMiner.getServicesDescription(services[i]);
	        if (txtDesc != null)
	        {
		        if (txtDesc.getResource() instanceof Resource)
		        {
		            hotelRes.addProperty(txtDesc.getProperty(), 
		                    txtDesc.getResource());
		        }
		        if (txtDesc.getResource() instanceof Resource[])
		        {
		            Resource[] resArr = (Resource[])txtDesc.getResource();
		            for (int j = 0; j < resArr.length; j++)
		                hotelRes.addProperty(txtDesc.getProperty(), 
			                    resArr[j]);
		        }
	        }
	    }
	}
	
	private void extractHotelAddress(Document hotelPage, HotelInfo hotelInfo)
		throws WebSourceParseException
	{
	    if (hotelPage == null)
	        throw new IllegalArgumentException("Document to parse can't be null");
	
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//DIV[@id='headlineContainer']/DIV[1]/H1/text()");
		
			Node node = (Node)xpath.selectSingleNode(hotelPage);
			if (node == null) return;
			hotelInfo.setName(node.getNodeValue());
			
			xpath = 
		        new DOMXPath("//DIV[@id='addressContainer']/DIV[@id='address']" +
		        "/text()");
			List nodes = xpath.selectNodes(hotelPage);
			Iterator nodesIter = nodes.iterator();
			String nodeValue;
			int idx = 0;
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    parseHotelAddress(node.getNodeValue(), idx++, hotelInfo);
			}
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelPage.getDocumentURI());
		}
	}
	
	private void parseHotelAddress(String nodeValue, int idx, HotelInfo hotelInfo)
	{
	    nodeValue = nodeValue.replaceAll("\n", "");
	    nodeValue = nodeValue.replaceAll("\t", "");
	    nodeValue = nodeValue.trim();
	    
	    if (idx == 0) //street
	    {
		    hotelInfo.setStreet(nodeValue);
	    }
	    else if (idx == 1) //locality, pcode
	    {
	        hotelInfo.setLocality(nodeValue);
	    }
	    else if (idx == 2) //country
	    {
	       hotelInfo.setCountry(nodeValue);
	    }
	    else if (idx == 3) //phone
	    {
	        nodeValue = nodeValue.substring(6, nodeValue.length() - 1);
	        hotelInfo.setPhone(nodeValue);
	    }
	    else if (idx == 4) //fax
	    {
	        nodeValue = nodeValue.substring(4, nodeValue.length() - 1);
	        hotelInfo.setFax(nodeValue);
	    }
	}
	
	private void extractCheckInCheckOutInfo(Document hotelDoc, Resource hotelRes)
		throws WebSourceParseException
	{
	    try
		{
		    DOMXPath xpath = 
		        new DOMXPath("//DIV[@id='checkInCheckOut']/" +
		                "descendant::B/text()");
		
			List nodes = xpath.selectNodes(hotelDoc);
			Iterator nodesIter = nodes.iterator();    
			Node node; 
			String nodeValue;
			StringBuffer checkBuffer = new StringBuffer();
			while (nodesIter.hasNext())
			{
			    node = (Node)nodesIter.next();
			    nodeValue = node.getNodeValue();;
			    
			    checkBuffer.append(nodeValue);
			    checkBuffer.append("; ");
			}
			
			String checkInOut = checkBuffer.toString().trim();
			
			Resource checkRes = hotelModel.createResource(
		            hotelRes.getURI() + "/CheckIn-CheckOut");
		    checkRes.addProperty(Hotel.additionalDetail, checkInOut);
		    hotelRes.addProperty(Hotel.additionalDetail, checkRes);
		}
		catch (Exception ex)
		{
		    ex.printStackTrace();
		    throw new WebSourceParseException(myAgent.getAID(), 
			        "extractHotelAddress() failed", ex, 
			        hotelDoc.getDocumentURI());
		}
	}
}
