package semfly.agents.behaviours;

import jade.lang.acl.ACLMessage;

import java.util.Enumeration;
import java.util.Vector;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class plays as a business logic for WrapperSender behaviour.
 */
public class WrapperSenderBL
{
    /**
	 * Evaluates received proposals. 
	 * The best proposer is the one with the lowest number of tokens in its TPQ. 
	 */
    public ACLMessage evaluateProposals(Vector responses, Vector acceptances)
    {
        if (responses == null)
            throw new IllegalArgumentException("Responses can't be null");
        if (acceptances == null)
            throw new IllegalArgumentException("Acceptances can't be null");
        
        ACLMessage accept = selectBestProposition(responses);
        return createReplies(responses, acceptances, accept);
    }
    
    /**
     * Returns the best proposition.
     */
    public ACLMessage selectBestProposition(Vector responses)
    {
        ACLMessage accept = null;
		Enumeration e = responses.elements();
		int minNoT = 0, noT = 0;
		boolean minSet = false;
		
		while (e.hasMoreElements()) 
		{
			ACLMessage msg = (ACLMessage) e.nextElement();
			if (msg.getPerformative() == ACLMessage.PROPOSE) 
			{
			    noT = extractNumberOfTokens(msg);
			    if (minSet)
			    {
					if (minNoT > noT) 
					{
					    minNoT = noT;
					    accept = msg;
					}
			    }
			    else
			    {
			        minNoT = noT;
			        accept = msg;
			        minSet = true;
			    }
			        
			}
		}
		
		return accept;
    }
    
    /**
     * Creates replies for the propositions received.
     * Returns reply for the accepted proposition.
     */
    private ACLMessage createReplies(Vector responses, Vector acceptances, 
            ACLMessage acceptedMsg)
    {
        ACLMessage reply, retReply = null;
		Enumeration e = responses.elements();
		while (e.hasMoreElements()) 
		{
			ACLMessage msg = (ACLMessage) e.nextElement();
			if (msg.getPerformative() == ACLMessage.PROPOSE) 
			{	
			    if (msg != acceptedMsg)
			    {
			        reply = msg.createReply();
					reply.setPerformative(ACLMessage.REJECT_PROPOSAL);
					acceptances.addElement(reply);
			    }
				else
				{
				    retReply = reply = msg.createReply();
				    reply.setPerformative(ACLMessage.ACCEPT_PROPOSAL);
					acceptances.addElement(reply);
				}
			}
		}
		
		return retReply;
    }
    
    /**
	 * Extracts a number of tokens (IndexToken) from a content of the ACL message
	 * given.
	 */
	public int extractNumberOfTokens(ACLMessage msg)
	{
	    if (msg == null)
	        throw new IllegalArgumentException("Message can't be null");
	    
	    return Integer.parseInt(msg.getContent());
	    
	}
}
