package semfly.db;

import java.io.StringReader;
import java.sql.SQLException;
import java.util.Date;

import semfly.exceptions.ExceptionManager;
import semfly.onto.jade.IndexToken;
import semfly.onto.jena.AdministrativeInfo;

import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.db.ModelRDB;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
//import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.shared.JenaException;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class is responsible for writing data from an IndexToken into a DB.
 */
public class DBWriter
{
    protected IDBConnection dbConn;
	protected String dbModelName;
	protected ModelRDB dbModel;
	protected IndexToken idxToken;
	
	protected DBWriter()
	{
	    
	}
	
	public DBWriter(IDBConnection dbConnection, String dbMdlName)
	{
	    assert dbConnection != null;
	    
	    dbConn = dbConnection;
	    dbModelName = dbMdlName;
	}
	
	/**
	 * Saves the IndexToken (indexToken) given to DB.
	 * 
	 * @param indexToken
	 * @throws JenaException
	 * @throws IllegalArgumentException
	 */
	public void write2DB(IndexToken indexToken) throws JenaException, 
		IllegalArgumentException
	{
	    assertIndexToken(indexToken);
		
		idxToken = indexToken;
		
		dbModel = createDBModel();
		Model inModel = createModel(idxToken.getJenaModel());
		
		writeAdminProperties(inModel);
		dbModel.add(inModel);
	}
	
	/**
	 * Saves administrative properties for the model given in a DB.
	 * 
	 * @param inModel Input model.
	 */
	protected void writeAdminProperties(Model inModel)
	{
	    Resource inSubject, exSubject;
		ResIterator subjectsIter = inModel.listSubjects();
		boolean justCreated = false, justUpdated = true;
		while (subjectsIter.hasNext())
		{
		    inSubject = subjectsIter.nextResource();
		    if (inSubject.isAnon()) continue;
			
		    justCreated = !dbModel.contains(inSubject, null);
			exSubject = dbModel.createResource(inSubject.getURI());
			
			StmtIterator properties = inSubject.listProperties();
			justUpdated = !dbModel.containsAll(inSubject.listProperties());
			addAdminProperties(exSubject, idxToken, justCreated, justUpdated);
		}
	}
	
	/**
	 * Adds administrative properties.
	 */
	private void addAdminProperties(Resource subject, IndexToken indexToken,
	        boolean justCreated, boolean justModified)
	{
		assert subject != null;
		
		String creator = indexToken.getCreatedBy().getName();
		Date creationDate = indexToken.getCreatedDate();
		Statement statement;
		
		if (justCreated)
		{
			subject.addProperty(AdministrativeInfo.createdBy, creator);
			subject.addProperty(AdministrativeInfo.createdDate, creationDate);
		}
	
		if (justModified)
		{
			statement = subject.getProperty(AdministrativeInfo.modifiedBy);
			if (statement != null) statement.changeObject(creator);
			else subject.addProperty(AdministrativeInfo.modifiedBy, creator);
			
			statement = subject.getProperty(AdministrativeInfo.modifiedDate);
			if (statement != null) statement.changeObject(creationDate);
			else subject.addProperty(AdministrativeInfo.modifiedDate, creationDate);
		}
		
		statement = subject.getProperty(AdministrativeInfo.modificationTryBy);
		if (statement != null) statement.changeObject(creator);
		else subject.addProperty(AdministrativeInfo.modificationTryBy, creator);
		
		statement = subject.getProperty(AdministrativeInfo.modificationTryDate);
		if (statement != null) statement.changeObject(creationDate);
		else subject.addProperty(AdministrativeInfo.modificationTryDate, creationDate);
	}
	
	/**
	 * Creates a DB model.
	 * @return Created model.
	 */
	protected ModelRDB createDBModel()
	{
	    ModelRDB model;
        
        if( !dbConn.containsModel(dbModelName) )
        	model = ModelRDB.createModel(dbConn, dbModelName);
        else
        	model = ModelRDB.open(dbConn, dbModelName);
        
        return model;
	}
	
	/**
	 * Creates model from the string given.
	 * @param strModel Model.
	 * @return Created model.
	 */
	private Model createModel(String strModel)
	{
		StringReader reader = new StringReader(strModel);	
		return ModelFactory.createDefaultModel().read(reader, "");
	}

	/**
	 * Checks if the property (prop) given is an administrative one.
	 * @param prop Property to check.
	 * @return True - administrative property, otherwise False.
	 */
	private boolean isAdminProperty(Property prop)
	{
		assert prop != null;
		
		return prop.getNameSpace().equals(AdministrativeInfo.getURI());
	}
	
	/**
	 * Checks if the token given is correctly build.
	 * @param token
	 * @throws IllegalArgumentException
	 */
	protected void assertIndexToken(IndexToken token) 
		throws IllegalArgumentException
	{
	    assert token != null;
	    
	    if (token.getCreatedBy() == null || 
	        token.getCreatedDate() == null ||
	        token.getJenaModel() == null
	        )
	    {
	        throw new 
	        	IllegalArgumentException("IndexToken isn't correctly built");
	    }
	}

	/**
	 * Closes a db connection.
	 */
	public void dispose()
	{
	    if (dbConn != null)
	    {
	        try { dbConn.close(); }
	        catch (SQLException ex) { ExceptionManager.publish(ex); }
	        finally { dbConn = null; }
	    }
	}
	
	/**
	 * Sets a DB model name.s
	 * @param dbModelName
	 */
	protected void setDBModelName(String dbModelName)
    {
        this.dbModelName = dbModelName;
    }
}
