package semfly.globals;

import jade.core.AID;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import semfly.exceptions.ExceptionManager;
import jade.core.Agent;

/**
 * @author Szymon Pisarek (szymekpi@o2.pl)
 *
 * This class is responsible for communication with Directory Facilator 
 * - JADE Agent. It is used as searching for agents.
 * 
 * Singleton Patter is used to build this class.
 */
public class DFAdapter
{
    private static DFAdapter searcher = null;
    
    protected DFAdapter()
    {
        
    }
    
    /**
     * Returns sole instance of DFSearcher class. If there is no instance
     * of this class, creates it.
     * 
     * @return The sole instance of this class.
     */
    public static DFAdapter getInstance()
    {
        if (searcher == null) searcher = new DFAdapter();
        
        return searcher;
    }
    
    /**
	 * Search all agents based on the service type given.
	 * @param agent The agent that is performing the search.
	 * @param serviceType Service type.
	 * @return AIDs of found agents of the service given, null if no agent 
	 * has been found or an exception has occured.
	 */
    public AID[] searchDF(Agent agent, String serviceType)
	{
        if (agent == null)
	        throw new IllegalArgumentException("Agent can't be null");
        
		DFAgentDescription dfd = new DFAgentDescription();
   		ServiceDescription sd = new ServiceDescription();
   		sd.setType(serviceType);
		dfd.addServices(sd);
		
		SearchConstraints ALL = new SearchConstraints();
		ALL.setMaxResults(new Long(-1));

		try
		{
			DFAgentDescription[] result = DFService.search(agent, dfd, ALL);
			AID[] agents = new AID[result.length];
			
			for (int i = 0; i < result.length; i++) 
				agents[i] = result[i].getName();
			
			return agents;
		}
        catch (FIPAException ex) 
		{ 
        	ExceptionManager.publish(ex);
        }
        
      	return null;
	}
    
    /**
	 * Registeres an agent in DF service.
	 * @param agent The agent that is performing the registration.
	 * @param sd Description of a service Agent wants to register
	 */
	public void register(Agent agent, ServiceDescription sd)
	{
	    assert agent != null;
	    
		DFAgentDescription dfd = new DFAgentDescription();
		dfd.setName(agent.getAID());

        try {
        	DFAgentDescription list[] = DFService.search(agent, dfd);
			if ( list.length>0 ) 
            	DFService.deregister(agent);
            	
            dfd.addServices(sd);
			DFService.register(agent,dfd);
		}
        catch (FIPAException ex) 
		{ 
        	ExceptionManager.publish(ex); 
        }
	}
}
